class Base:
    name = ''
    symbol = ''
    factor = 1


class Yotta( Base ):
    name = 'yotta'
    symbol = 'Y'
    factor = 10 ** 24


class Zetta( Base ):
    name = 'zetta'
    symbol = 'Z'
    factor = 10 ** 21


class Exa( Base ):
    name = 'exa'
    symbol = 'E'
    factor = 10 ** 18


class Peta( Base ):
    name = 'peta'
    symbol = 'P'
    factor = 10 ** 15


class Tera( Base ):
    name = 'tera'
    symbol = 'T'
    factor = 10 ** 12


class Giga( Base ):
    name = 'giga'
    symbol = 'G'
    factor = 10 ** 9


class Mega( Base ):
    name = 'mega'
    symbol = 'M'
    factor = 10 ** 6


class Kilo( Base ):
    name = 'kilo'
    symbol = 'k'
    factor = 10 ** 3


class Hecto( Base ):
    name = 'hecto'
    symbol = 'h'
    factor = 10 ** 2


class Deca( Base ):
    name = 'deca'
    symbol = 'da'
    factor = 10 ** 1


class Centi( Base ):
    name = 'Centi'
    symbol = 'c'
    factor = 10 ** -2


class Mili( Base ):
    name = 'mili'
    symbol = 'm'
    factor = 10 ** -3


class Micro( Base ):
    name = 'micro'
    symbol = 'μ'
    factor = 10 ** -6


class Nano( Base ):
    name = 'nano'
    symbol = 'n'
    factor = 10 ** -9


class Pico( Base ):
    name = 'pico'
    symbol = 'p'
    factor = 10 ** -12


class Femto( Base ):
    name = 'femto'
    symbol = 'f'
    factor = 10 ** -15


class Atto( Base ):
    name = 'Atto'
    symbol = 'a'
    factor = 10 ** -18


class Zepto( Base ):
    name = 'zepto'
    symbol = 'z'
    factor = 10 ** -21


class Yocto( Base ):
    name = 'yocto'
    symbol = 'y'
    factor = 10 ** -24



import importlib


def import_( module, cls=None ):
    if cls is None:
        split = module.split( '.' )
        cls = split[-1]
        module = ".".join( split[:-1] )
        return import_( module=module, cls=cls )
    module = importlib.import_module( module )
    return getattr( module, cls )


def export( cls ):
    if isinstance( cls, type ):
        return "{module}.{cls}".format(
            module=cls.__module__, cls=cls.__name__ )
    else:
        return export( cls=cls.__class__ )



from unittest import TestCase

from chibi.file.snippets import get_relative_path


class Test_get_relative_path( TestCase ):
    def test_should_return_the_relative_paths( self ):
        result = get_relative_path( '/usr/var/log', '/usr/var/security' )
        self.assertEqual( [ 'log', 'security' ], result  )

    def test_should_return_the_relative_path( self ):
        result = get_relative_path( '/usr/var/log', root='/usr/var' )
        self.assertEqual( 'log', result  )



from unittest import TestCase

from chibi.object import Chibi_object
from chibi.object.descriptor import (
    String, Dict, Tree_simple, Dict_defaults, Set
)


class Chibi_object_empty( Chibi_object ):
    pass


class Chibi_object_with_descriptors( Chibi_object ):
    name = String()
    test_dict = Dict()
    test_dict_default = Dict_defaults()
    test_tree = Tree_simple()
    test_set = Set()


class Chibi_object_with_defaults( Chibi_object ):
    name = String( default='hello' )
    test_dict_default = Dict_defaults( default='word' )


class Test_chibi_object( TestCase ):

    def test_chibi_object_simple( self ):
        obj = Chibi_object_empty()
        self.assertIsInstance( obj, Chibi_object )

    def test_with_descriptors( self ):
        obj = Chibi_object_with_descriptors()
        self.assertIsInstance( obj, Chibi_object )
        self.assertEqual( obj.name, '' )

        obj.name = 'hellooo'
        self.assertEqual( obj.name, 'hellooo' )
        self.assertIsNone( obj.test_dict )
        obj.test_dict = { 'key': 'test' }
        self.assertEqual( obj.test_dict, { 'key': 'test' } )

        self.assertFalse( obj.test_dict_default )
        data = obj.test_dict_default[ 'sadf' ]
        self.assertIsNone( data )

        self.assertFalse( obj.test_tree )
        self.assertEqual( list( obj.test_tree.keys() ), [] )

        self.assertEqual( len( obj.test_set ), 0 )

    def test_with_descriptior_assing( self ):
        obj = Chibi_object_with_descriptors( name='stuff', test_dict={} )
        self.assertEqual( obj.test_dict, {} )
        self.assertEqual( obj.name, 'stuff' )
        obj.name = 'asdf'
        self.assertEqual( obj.name, 'asdf' )
        obj.test_dict['asdf'] = 123
        self.assertEqual( obj.test_dict, { 'asdf': 123 } )

        self.assertFalse( obj.test_tree )
        obj.test_tree.a.b.c
        self.assertEqual( obj.test_tree, { 'a': { 'b': { 'c': {} } } } )
        obj.test_tree.a.rusky = 'RUSH B'
        self.assertEqual( obj.test_tree, { 'a': { 'rusky': 'RUSH B',
                                                  'b': { 'c': {} } } } )

        obj.test_dict_default[ 'qwer' ] = 'word'
        self.assertIsNotNone( obj.test_dict_default[ 'qwer' ] )
        self.assertEqual( obj.test_dict_default[ 'qwer' ], 'word' )

        obj.test_set |= set( 'abc' )
        self.assertEqual( len( obj.test_set ), 3 )

    def test_with_defaults( self ):
        obj = Chibi_object_with_defaults()
        self.assertEqual( obj.name, 'hello' )
        obj.name = 'zxcv'
        self.assertEqual( obj.name, 'zxcv' )

        word = obj.test_dict_default[ 'hello' ]
        self.assertEqual( word, 'word' )



import random

from faker import Factory as Faker_factory

from chibi.file.snippets import move, exists, ls, get_file_from_path
from tests.snippet.files import Test_with_files


faker = Faker_factory.create()


class Test_move( Test_with_files ):
    def test_when_move_a_empty_file_should_create_a_new_empty_file( self ):
        file = random.choice( self.files )
        dest = self.root_dir + faker.file_name()
        self.assertFalse( dest.exists )
        with open( file ) as file_str:
            self.assertFalse( file_str.read() )

        move( file, dest )
        self.assertFalse( exists( file ) )
        with open( dest ) as file_dest:
            self.assertFalse( file_dest.read() )

    def test_when_move_a_file_with_content_should_have_the_content( self ):
        file = random.choice( self.files_with_content )
        dest = self.root_dir + faker.file_name()
        self.assertFalse( dest.exists )
        with open( file ) as file_src:
            str_data = file_src.read()

        move( file, dest )
        self.assertFalse( exists( file ) )
        with open( dest ) as file_dest:
            self.assertEqual( str_data, file_dest.read() )

    def test_when_move_with_wild_card_should_move_all_the_files( self ):
        self.assertTrue( list( ls( self.folder_with_files_with_content ) ) )
        files = list( ls( self.folder_with_files_with_content ) )
        source = self.folder_with_files_with_content + '*'
        move( source, self.root_dir )
        for f in files:
            self.assertFalse( exists( f ) )

        files = [ get_file_from_path( f ) for f in files ]

        for f in files:
            self.assertTrue( ( self.root_dir + f ).exists )



from chibi.file.snippets import ls_only_dir, is_dir
from tests.snippet.files import Test_with_files


class Test_ls_only_dir( Test_with_files ):
    def test_should_list_all_dirs_and_not_the_files_from_root( self ):
        result = list( ls_only_dir( self.root_dir ) )
        self.assertTrue( result, "cannot find dirs in the root dir" )
        for dir in result:
            self.assertTrue( ( self.root_dir + dir ).is_a_folder )

    def test_should_return_a_empty_list( self ):
        result = list( ls_only_dir( self.empty_folder ) )
        self.assertFalse( result )

    def test_when_is_not_the_current_folder_should_get_the_dirs( self ):
        for dir_level_1 in self.dirs:
            directories_in_sub_folder = list( ls_only_dir( dir_level_1 ) )
            self.assertTrue(
                directories_in_sub_folder,
                "cannot find dirs in the the dir: {}".format(
                    directories_in_sub_folder ) )

            for dir in directories_in_sub_folder:
                abs_dir = dir_level_1 + dir
                self.assertTrue(
                    is_dir( abs_dir ),
                    "is was find the dir {} but is not a dir".format(
                        abs_dir ) )



from unittest import TestCase

from chibi.atlas import loads, Chibi_atlas


class Test_loads_json( TestCase ):
    def setUp( self ):
        self.string = """
        { "tests": { "test_1": "123" } }
        """

    def test_should_return_a_chibi_atlas( self ):
        result = loads( self.string )
        self.assertIsInstance( result, Chibi_atlas )
        self.assertTrue( result )


class Test_loads_yaml( Test_loads_json ):
    def setUp( self ):
        self.string = """martin:
    name: Martin D'vloper
    job: Developer
    skills:
    - python
    - perl
    - pascal
tabitha:
    name: Tabitha Bitumen
    job: Developer
    skills:
    - lisp
    - fortran
    - erlang

        """


class Test_loads_xml( Test_loads_json ):
    def setUp( self ):
        self.string = """<?xml version="1.0" encoding="UTF-8"?>
        <note>
            <to>Tove</to>
            <from>Jani</from>
            <heading>Reminder</heading>
            <body>Don't forget me this weekend!</body>
        </note>
        """



import tempfile
from chibi.file import Chibi_path

tmp_folder = ''


def make_empty_file( suffix=None ):
    folder = get_temp_folder()
    return Chibi_path( tempfile.mkstemp(
        dir=str( folder ), suffix=suffix )[1] )


def get_temp_folder():
    global tmp_folder
    if not tmp_folder:
        tmp_folder = tempfile.mkdtemp()
    return Chibi_path( tmp_folder )



from chibi.atlas import Chibi_atlas


def group_by( l, key ):
    """
    agrupa por alguna llave

    Arguments
    =========
    l: list, tuple, finite iter
    key: callable or str
        funcion que separara los elementos
        si es un string se asume que los elementos son dicionarios
        y el valor de key se usara para obtener el valor

    Return:
        Chibi_atlas
            la llave es el valor del key y el valor es una lista
            de con los elementos de ese key
    """
    if isinstance( key, str ):
        return _group_by_str( l, key )
    elif callable( key ):
        return _group_by_callable( l, key )
    else:
        raise NotImplementedError(
            f'no esta implementado si la llave es un tipo {type( key )}' )


def _group_by_str( l, key ):
    result = Chibi_atlas()
    for item in l:
        value = item[ key ]
        try:
            result[ value ].append( item )
        except KeyError:
            result[ value ] = [ item ]
    return result


def _group_by_callable( l, key ):
    result = Chibi_atlas()
    for item in l:
        value = key( item )
        try:
            result[ value ].append( item )
        except KeyError:
            result[ value ] = [ item ]
    return result



from chibi.file.path import Chibi_path
from chibi.file.snippets import add_extensions
from tests.snippet.files import Test_with_files


class Test_add_extensions( Test_with_files ):
    def test_should_add_thumbnail( self ):
        result = add_extensions( "image.jpg", "thumbnail" )
        expected = "image.thumbnail.jpg"
        self.assertEqual( result, expected )

    def test_should_return_a_chibi_path( self ):
        result = add_extensions( "image.jpg", "thumbnail" )
        self.assertIsInstance( result, Chibi_path )



from tests.snippet.files import Test_with_files
from chibi.file import Chibi_file


class Test_find( Test_with_files ):
    def setUp( self ):
        super().setUp()
        self.chibi_file = Chibi_file( self.files_with_content[0] )

    def test_should_no_fail( self ):
        result = self.chibi_file.find( '' )
        self.assertFalse( result )

    def test_should_find_the_text( self ):
        self.chibi_file.write( 'Pipiru-piru-piru-pipiru-pii' )
        result = self.chibi_file.find( 'piru' )
        self.assertTrue( result )
        self.assertTrue( 'piru' in self.chibi_file )

    def test_should_no_find_the_text( self ):
        self.chibi_file.write( 'Pipiru-piru-piru-pipiru-pii' )
        result = self.chibi_file.find( 'poro' )
        self.assertTrue( result == -1 )
        self.assertFalse( 'poro' in self.chibi_file )



from chibi.file import Chibi_file
from chibi.file.snippets import add_extensions
from chibi.file.path import Chibi_path
from PIL import Image


class Chibi_image( Chibi_file ):
    @property
    def dimension( self ):
        return self._PIL.size

    @property
    def _PIL( self ):
        return Image.open( self.path )

    def read_cv2( self ):
        import cv2 as cv
        return cv.imread( self.path, cv.IMREAD_COLOR )

    def read_cv( self ):
        import cv2 as cv
        return cv.imread( self.path, cv.IMREAD_COLOR )

    def read_cv_gray( self ):
        import cv2 as cv
        return cv.imread( self.path, cv.IMREAD_GRAYSCALE )

    def __eq__( self, other ):
        if not isinstance( other, Chibi_image ):
            return False
        return (
            self.properties.mime == other.properties.mime
            and self.dimension == other.dimension
            and self.properties.size == other.properties.size
        )

    def thumbnail( self, path, size=( 64, 64 ) ):
        path = Chibi_path( path )
        if ( path.is_a_folder ):
            path = path + add_extensions( self.file_name, 'thumbnail' )
            thumbnail = self._PIL.copy()
            thumbnail.thumbnail( size )
            thumbnail.save( path )
            return type( self )( path )
        else:
            raise NotImplementedError

    def show( self ):
        return self._PIL.show()

    '''
    @property
    def flatter( self, h, w ):
        import cv2 as cv
        row = cv.resize(
            image,( h, w ), interpolation=cv.INTER_AREA ).flatten()
        col = cv.resize(
            image,( h, w ), interpolation=cv2.INTER_AREA ).flatten( 'F' )
        return col, row
    '''



import itertools


def chunks( iterable, size=1000 ):
    """
    Make digestible chunks of something that can be iterated

    Arguments
    ---------
    iterable: list, tuple, generator
        Just have to support the iter function will break it
        in digestible chunks
    size: size of the digestibles chunks

    Examples
    --------
    >>> [ list( chunk ) for chunk in chunks( range( 10 ), size=5 ) ]
    [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]
    """
    iterator = iter( iterable )
    for i in iterator:
        yield itertools.chain( [i], itertools.islice( iterator, size - 1 ) )


class ___DUMP:
    def __init__( self, dump ):
        self.dump = dump


def chunk_each( l, f ):
    inner_result = []
    iterator = iter( l )
    for i in iterator:
        if f( i ):
            if inner_result:
                yield inner_result
            inner_result = []
        inner_result.append( i )
    if inner_result:
        yield inner_result


def cut_until( iterator, until ):
    for i in iterator:
        if until( i ):
            break
        yield i



from unittest import TestCase
from chibi.chain import Chibi_chain


class Test_chibi_chain( TestCase ):
    def test_should_instance_like_a_normal_list( self ):
        l = Chibi_chain()
        self.assertFalse( l )
        l = Chibi_chain( range( 10 ) )
        self.assertEqual( l, list( range( 10 ) ) )
        l = Chibi_chain( [ 0, 1, 2, ] )
        self.assertEqual( l, list( range( 3 ) ) )

    def test_when_no_has_next_object_the_iter_should_work_like_a_list( self ):
        l = Chibi_chain( range( 2 ) )
        i = iter( l )
        self.assertEqual( next( i ), 0 )
        self.assertEqual( next( i ), 1 )
        with self.assertRaises( StopIteration ):
            self.assertEqual( next( i ), 2 )

    def test_when_have_next_object_should_concatenate_the_object( self ):
        l = Chibi_chain(
            range( 2 ), next_obj=range( 10, 12 ),
            retrieve_next=lambda obj: list( obj ) )
        i = iter( l )
        self.assertEqual( next( i ), 0 )
        self.assertEqual( next( i ), 1 )
        self.assertEqual( next( i ), 10 )
        self.assertEqual( next( i ), 11 )
        with self.assertRaises( StopIteration ):
            next( i )
        self.assertEqual( list( l ), [ 0, 1, 10, 11 ] )

    def test_when_the_next_obj_is_a_chibi_chain_should_concatenate( self ):
        n = Chibi_chain( range( 10, 12 ) )
        l = Chibi_chain(
            range( 2 ), next_obj=n, retrieve_next=lambda obj: obj )
        i = iter( l )
        self.assertEqual( next( i ), 0 )
        self.assertEqual( next( i ), 1 )
        self.assertEqual( next( i ), 10 )
        self.assertEqual( next( i ), 11 )
        with self.assertRaises( StopIteration ):
            next( i )
        self.assertEqual( list( l ), [ 0, 1, 10, 11 ] )

    def test_when_have_a_chain_of_chain_should_concatenate_all( self ):
        p = Chibi_chain( range( 20, 22 ) )
        n = Chibi_chain(
            range( 10, 12 ), next_obj=p, retrieve_next=lambda obj: obj )
        l = Chibi_chain(
            range( 2 ), next_obj=n, retrieve_next=lambda obj: obj )
        i = iter( l )
        self.assertEqual( next( i ), 0 )
        self.assertEqual( next( i ), 1 )
        self.assertEqual( next( i ), 10 )
        self.assertEqual( next( i ), 11 )
        self.assertEqual( next( i ), 20 )
        self.assertEqual( next( i ), 21 )
        with self.assertRaises( StopIteration ):
            next( i )
        self.assertEqual( list( l ), [ 0, 1, 10, 11, 20, 21 ] )



from dateutil.parser import parse


def is_date( string ):
    """
    check is a string is a date

    Arguments
    ---------
    string: str
        string to evaluate if is a date

    Returns
    -------
    data, datetime
        if the date is a date or datetime format
    bool
        if the string is not a date return False
    """
    try:
        return parse( string )
    except ValueError:
        return False


def is_iter( obj ):
    """
    evaluate if the object can be iterate

    Arguments
    ---------
    obj: object

    Returns
    -------
    bool
    """
    try:
        iter( obj )
        return True
    except TypeError:
        return False


def is_like_list( obj ):
    """
    evaluate if the object is like a list

    Arguments
    ---------
    obj: object

    Returns
    -------
    bool
    """
    return isinstance( obj, ( list, tuple, set ) )



import json
import re

from . import regex as chibi_regex
from .string import camel_to_snake
from chibi.snippet.is_type import is_like_list


def keys_to_snake_case( d ):
    """
    transforma los keys de un dicionaria a snake case

    Returns
    =======
    dict

    Examples
    ========
    >>>keys_to_snake_case( { "HolaMundo": "hola mundo" } )
    { "hola_mundo": "hola_mundo" }
    """
    result = {}
    for k, v in d.items():
        if isinstance( k, str ):
            k = camel_to_snake( k )
        if isinstance( v, ( dict, list, tuple ) ):
            v = __inner_keys_to_snake_case( v )
        result[k] = v
    return result


def replace_keys( d, dr ):
    """
    replace the the keys of a dictionary using another dictioanry for guide

    Parameters
    ----------
    d: dict
        dict is going to have replace his keys
    dr: dict
        dict is going to use for remplace his keys

    Examples
    ========
    >>>replace_keys( { 'a': 'a' }, { 'a': 'b' } )
    {'b':'a'}
    """
    if is_like_list( d ):
        for v in d:
            replace_keys( v, dr )
    elif isinstance( d, dict ):
        keys = tuple( d.keys() )
        for k in keys:
            try:
                new_key = dr[k]
                d[ new_key ] = d[k]
                del d[k]
                k = new_key
            except KeyError:
                pass
            replace_keys( d[k], dr )


def rename_keys( d, func ):
    """
    rename the keys in a dict using a function

    Parameters
    ==========
    d: dict
        target dict
    func: callable
        function is going to rename rename the keys

    Returns
    =======
    dict

    Examples
    ========
    >>>rename_keys( { 'a': 'a' }, lambda k: k.upper() )
    {'A':'a'}
    """
    if not callable( func ):
        raise NotImplementedError
    result = {}
    for k, v in d.items():
        result[ func( k ) ] = v
        if isinstance( v, dict ):
            rename_keys( v, func )
    return result


def lower_keys( d ):
    """
    lower all the keys in the dict

    Parameters
    ==========
    d: dict
        target dict

    Returns
    =======
    dict

    Examples
    ========
    >>>lower_keys( { 'A': 'a' } )
    {'a':'a'}
    """
    return rename_keys( d, func=lambda x: x.lower() )


def pop_regex( d, regex ):
    """
    do pop to the keys match with the regex and form a new dict with
    those keys and values

    parameters
    ==========
    d: dict
        target dict
    regex: str or regex
        string to compile the regex or a regex object

    Returns
    =======
    dict

    Examples
    ========
    >>>origin = { "a": "a", "b": "b", "aa": "aa" }
    >>>pop_regex( origin, r'a*' )
    { "a": "a", "aa": "aa" }
    >>>origin == { "b": "b" }
    True
    """
    if isinstance( regex, str ):
        regex = re.compile( regex )
    keys = [ k for k in d.keys() ]
    result = {}
    for k in keys:
        if chibi_regex.test( regex, k ):
            result[ k ] = d.pop( k )
    return result


def get_regex( d, regex ):
    """
    hace get a las llaves del dicionario que concuerden con el regex
    y forma un nuevo dicionario con ese

    parameters
    ==========
    d: dict
        dicionarrio que se le haran los gets
    regex: str or regex
        string que se compilara para ser regex

    Examples
    ========
    >>>origin = { "a": "a", "b": "b", "aa": "aa" }
    >>>pop_regex( origin, r'a*' )
    { "a": "a", "aa": "aa" }
    >>>origin == { "a": "a", "b": "b", "aa": "aa" }
    True
    """
    if isinstance( regex, str ):
        regex = re.compile( regex )
    keys = [ k for k in d.keys() ]
    result = {}
    for k in keys:
        if chibi_regex.test( regex, k ):
            result[ k ] = d.get( k )
    return result


def delete_list_of_keys( d, *keys ):
    """
    elimina las keys de un dicionario

    Parameters
    ==========
    d: dict
        dicionario del que se eliminaran las keys
    keys: tuple
        llaves a eliminar

    Examples
    ========
    >>>origin = { 'a': 'a', 'b': 'b': 'c': 'c' }
    >>>delete_list_of_keys( origin, 'b', 'c' )
    { 'a': 'a' }
    origin == { 'a': 'a' }
    """
    for key in keys:
        del d[ key ]
    return d


def get_list_of_keys( d, *keys ):
    """
    crea un nuevo dicionario con el subconjunto de llaves

    Parameters
    ==========
    d: dict
    keys: tuple

    Examples
    ========
    >>>origin = { 'a': 'a', 'b': 'b': 'c': 'c' }
    >>>get_list_of_keys( origin, 'b', 'c' )
    { 'b': 'b': 'c': 'c' }
    """
    return { key: d[ key ] for key in keys }


def get_from_dict( d, **kw ):
    """
    crea un nuevo dicionario usando el kw
    la llave del kw es la llave del dicionario
    y el valor de kw es el nombre de la nueva llave

    Parameters
    ==========
    d: dict
    kw: dict

    Examples
    ========
    >>>origin = { 'a': 'a', 'b': 'b': 'c': 'c' }
    >>>get_from_dict( origin, a='d', c='e' )
    { 'd': 'a': 'e': 'c' }
    """
    return { v: d[ k ] for k, v in kw.items() }


def remove_value( d, element ):
    """
    elimina los items de un dicionario en el que su value es element

    Parameters
    ==========
    d: dict
    element: object

    Returns
    =======
    dict

    Examples
    ========
    >>>remove_value( { 'a': 1, 'b': 2, 'c': 1 }, 1 )
    { 'b': 2 }
    """
    keys_to_delete = []
    for key, value in d.items():
        if value is element:
            keys_to_delete.append( key )
            continue
        if isinstance( value, dict ):
            r = remove_value( value, element )
        elif isinstance( value, list ):
            r = __remove_element__list( value, element )
        else:
            continue
        if r:
            d[ key ] = r
        else:
            keys_to_delete.append( key )

    delete_list_of_keys( d, *keys_to_delete )
    return d


def remove_nones( d ):
    """
    elimina los Nones de un dicionario

    Parameters
    ==========
    d: dict

    Returns
    =======
    dict

    Examples
    ========
    >>>remove_nones( { 'a': None, 'b': 2, 'c': None } )
    { 'b': 2 }
    """
    return remove_value( d, None )


def hate_ordered_dict( d ):
    """
    elimina los orderer dicts
    """
    return json.loads( json.dumps( d ) )


def remove_xml_notatation( d ):
    if isinstance( d, dict ):
        result = {}
        for k, v in d.items():
            if k.startswith( '#' ) or k.startswith( '@' ):
                k = k[1:]
            if ':' in k:
                k = k.replace( ':', '_' )
            result[ k ] = remove_xml_notatation( v )
        return result
    elif isinstance( d, list ):
        return [ remove_xml_notatation( i ) for i in d ]
    return d


def __remove_element__list( l, element ):
    result = []
    for i in l:
        if i is element:
            continue
        if isinstance( i, list ):
            result.append( __remove_element__list( i, element ) )
        if isinstance( i, dict ):
            result.append( remove_value( i, element ) )

    return result


def __inner_keys_to_snake_case( d ):
    if isinstance( d, dict ):
        return keys_to_snake_case( d )
    elif isinstance( d, list ):
        return [ __inner_keys_to_snake_case( a ) for a in d ]
    elif isinstance( d, tuple ):
        return tuple( __inner_keys_to_snake_case( a ) for a in d )
    return d


def split( d ):
    """
    divide el dicionario en varios dicionarios

    Params
    ======
    d: dict

    Results
    =======
    map: generador con los dicionarios

    Examples
    ========
    >>>list( split( { "a": "1", { "b": "2" } } ) )
    [ { "a": "1" }, { "b": "2" } ]
    """
    return map( lambda x: { x[0]: x[1] }, d.items() )


def group_by( d, key ):
    from chibi.atlas.multi import Chibi_atlas_multi
    result = Chibi_atlas_multi()
    if isinstance( d, dict ):
        try:
            result[ d[ key ] ] = d
        except KeyError:
            pass
    elif isinstance( d, list ):
        for dd in d:
            r = group_by( dd, key )
            result.update( r )
    return result



from chibi.file import Chibi_file
from tests.snippet.files import Test_with_files


class Test_chibi_file_properties( Test_with_files ):
    def setUp( self ):
        super().setUp()
        self.chibi_file = Chibi_file( self.files_with_content[0] )

    def test_file_name( self ):
        self.assertIn( self.chibi_file.file_name, self.chibi_file.path )
        self.assertNotEqual( self.chibi_file.file_name, self.chibi_file.path )

    def test_dir( self ):
        self.assertIn( self.chibi_file.dir, self.chibi_file.path )
        self.assertNotEqual( self.chibi_file.dir, self.chibi_file.path )



from .base import Unit
from .prefix import (
    Yotta, Zetta, Exa, Peta, Tera, Giga, Mega, Kilo, Hecto,
    Deca, Base, Centi, Mili, Micro, Nano, Pico, Femto, Atto,
    Zepto, Yocto )


class Metre( Unit ):
    name = 'metre'
    symbol = 'm'
    valid_prefix = (
        Yotta, Zetta, Exa, Peta, Tera, Giga, Mega, Kilo,
        Hecto, Deca, Base, Centi, Mili, Micro, Nano, Pico,
        Femto, Atto, Zepto, Yocto )


class Gram( Unit ):
    name = 'gram'
    symbol = 'g'
    valid_prefix = (
        Yotta, Zetta, Exa, Peta, Tera, Giga, Mega, Kilo,
        Hecto, Deca, Base, Centi, Mili, Micro, Nano, Pico,
        Femto, Atto, Zepto, Yocto )



from unittest import TestCase

from chibi.object.descriptor import Descriptor


class Test_descriptor( TestCase ):

    def test_descriptor_init( self ):
        d = Descriptor()
        self.assertIsNone( d.name )
        self.assertFalse( d.required )
        self.assertIsNone( d.default )

    def test_descriptor_default( self ):
        d = Descriptor( default='asdf' )
        self.assertIsNone( d.name )
        self.assertFalse( d.required )
        self.assertEqual( d.default, 'asdf' )



import os
import logging
from chibi.atlas import Chibi_atlas, Chibi_atlas_default
from chibi.file import Chibi_path
from chibi.file.other import Chibi_json, Chibi_yaml, Chibi_python
import chibi_donkey as donkey


logger = logging.getLogger( 'chibi.config.Configuration' )


__all__ = [ 'Configuration' ]


def _default_factory():
    return Configuration()


class Configuration( Chibi_atlas_default ):
    def __init__( self, default_factory=None, *args, **kw ):
        if default_factory is None:
            default_factory = _default_factory
        super().__init__( default_factory, *args, **kw )

    def load( self, path ):
        path = Chibi_path( path )
        with path as f:
            if isinstance( f, ( Chibi_json, Chibi_yaml ) ):
                for k, v in f.read().items():
                    self[ k ] = v
            elif isinstance( f, Chibi_python ):
                logger.info( f"ejecutanto archivo python {f}" )
                f.import_()
            else:
                raise NotImplementedError(
                    "no esta implementado la carga de configuracion de los "
                    f"archivos {type(f)} de {f}" )


class Logger_configuration( Chibi_atlas ):
    def __getitem__( self, name ):
        try:
            return super().__getitem__( name )
        except KeyError:
            # logger = logging.getLogger( name )
            self[ name ] = Logger( name=name )
            return super().__getitem__( name )


class Logger( Chibi_atlas ):
    @property
    def level( self ):
        logger = self.logger
        while logger:
            if logger.level != logging.NOTSET:
                return logger.level
            logger = logger.parent
        return self.logger.level

    @level.setter
    def level( self, value ):
        if isinstance( value, str ):
            level = getattr( logging, value )
            if isinstance( level, int ):
                self.logger.setLevel( level )
            else:
                raise NotImplementedError(
                    f"no esta implementado la asignacion del level de {value}"
                )
        elif isinstance( value, int ):
            self.logger.setLevel( value )
        else:
            raise NotImplementedError(
                f"no esta implementado la asignacion del level de {value}" )

    @property
    def logger( self ):
        return logging.getLogger( self.name )


class Env_vars( Configuration ):
    def __init__( self, default_factory=None, *args, **kw ):
        if default_factory is None:
            default_factory = str
        d = donkey.inflate( os.environ )
        super().__init__( default_factory, d, *args, **kw )



import random

from faker import Factory as Faker_factory

from chibi.file.snippets import delete, exists
from tests.snippet.files import Test_with_files


faker = Faker_factory.create()


class Test_delete( Test_with_files ):
    def test_when_delete_a_file_should_no_exists( self ):
        file = random.choice( self.files )
        self.assertTrue( exists( file ) )
        delete( file )
        self.assertFalse( exists( file ) )

    def test_whe_delete_a_dir_should_removed( self ):
        dir = random.choice( self.dirs )
        self.assertTrue( exists( dir ) )
        delete( dir )
        self.assertFalse( exists( dir ) )



import random

from faker import Factory as Faker_factory

from chibi.file.snippets import copy, exists, copy_folder, ls
from tests.snippet.files import Test_with_files


faker = Faker_factory.create()


class Test_copy( Test_with_files ):
    def test_when_copy_a_empty_file_should_create_a_new_empty_file( self ):
        file = random.choice( self.files )
        dest = self.root_dir + faker.file_name()
        self.assertFalse( exists( dest ) )

        copy( file, dest )
        self.assertTrue( exists( dest ) )
        with open( dest ) as file_dest:
            self.assertFalse( file_dest.read() )

    def test_when_copy_a_file_with_content_should_copy_the_content( self ):
        file = random.choice( self.files_with_content )
        dest = self.root_dir + faker.file_name()
        self.assertFalse( exists( dest ) )

        copy( file, dest )
        self.assertTrue( exists( dest ) )
        with open( dest ) as file_dest:
            with open( file ) as file_src:
                self.assertEqual( file_src.read(), file_dest.read() )


class Test_copy_folder( Test_with_files ):
    def test_should_copy_all_the_folder( self ):
        dest = self.root_dir + 'hola'
        self.assertFalse( exists( dest ) )
        copy_folder( self.folder_with_files_with_content, dest )
        self.assertTrue( exists( dest ) )

        self.assertEqual(
            len( set( ls( self.folder_with_files_with_content ) ) ),
            len( set( ls( dest ) ) ) )



from chibi.fancy.is_type import is_number
from chibi.units.prefix import Base


class Unit:
    name = 'abstract'
    symbol = 'abstract'
    valid_prefix = ( Base, )

    def __init__( self, value ):
        self.value = value
        self.prefix = Base

    def __str__( self ):
        return '{value}{prefix}{symbol}'.format(
            value=self.value * self.prefix.factor,
            symbol=self.symbol, prefix=self.prefix.symbol )

    def __repr__( self ):
        return '{cls}( {value} )'.format( str( self.__class__ ), str( self ) )

    def __add__( self, other ):
        if is_number( other ):
            return self.__class__( self.value + other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __sub__( self, other ):
        if is_number( other ):
            return self.__class__( self.value - other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __mul__( self, other ):
        if is_number( other ):
            return self.__class__( self.value * other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __truediv__( self, other ):
        if is_number( other ):
            return self.__class__( self.value / other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __floordiv__( self, other ):
        if is_number( other ):
            return self.__class__( self.value // other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __pow__( self, other ):
        if is_number( other ):
            return self.__class__( self.value ** other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __mod__( self, other ):
        if is_number( other ):
            return self.__class__( self.value % other )
        raise NotImplementedError(
            "cannot subtract no numbers, you send a {}".format(
                type( other ) ) )

    def __radd__( self, other ):
        return self + other

    def __rsub__( self, other ):
        return self - other

    def __rmul__( self, other ):
        return self * other

    def __rtruediv__( self, other ):
        return self / other

    def __rfloordiv__( self, other ):
        return self // other

    def __rpow__( self, other ):
        return self ** other



import chibi
import unittest


class Test_b64( unittest.TestCase ):
    def setUp( self ):
        self.encode = 'YXNkZg=='
        self.decode = 'asdf'

    def test_encode( self ):
        self.assertEqual( chibi.b64.encode( self.decode ), self.encode )

    def test_decode( self ):
        self.assertEqual( chibi.b64.decode( self.encode ), self.decode )



from unittest import TestCase
from chibi.snippet.list import group_by


class Test_group_by( TestCase ):
    def setUp( self ):
        self.l = [
            dict( s='ok' ),
            dict( s='no' ),
            dict( s='fail' ),
            dict( s='fail' ),
            dict( s='ok' ),
            dict( s='ok' ),
        ]
        self.expected = {
            'ok': [ dict( s='ok' ), dict( s='ok' ), dict( s='ok' ), ],
            'fail': [ dict( s='fail' ), dict( s='fail' ), ],
            'no': [ dict( s='no' ), ],
        }

    def test_should_return_expected_result_if_is_a_callable( self ):
        result = group_by( self.l, lambda v: v[ 's' ] )
        self.assertEqual( result, self.expected )

    def test_should_return_expected_result_if_is_a_string( self ):
        result = group_by( self.l, 's' )
        self.assertEqual( result, self.expected )



from chibi.file.snippets import exists
from tests.snippet.files import Test_with_files


class Test_exists( Test_with_files ):
    def test_should_be_true_with_a_file_exists( self ):
        for file in self.files:
            self.assertTrue( exists( file ) )

    def test_should_be_true_with_dirs( self ):
        self.assertTrue( exists( self.root_dir ) )
        for dir in self.dirs:
            self.assertTrue( exists( dir ) )

    def test_should_be_false_when_dont_exists_the_file( self ):
        self.assertFalse( exists( self.empty_folder + 'some_file' ) )



from unittest import TestCase

from faker import Factory as Faker_factory

from chibi.snippet.dict import (
    keys_to_snake_case, replace_keys, pop_regex, get_regex, rename_keys,
    lower_keys, delete_list_of_keys, get_list_of_keys, get_from_dict,
    remove_value, remove_nones, remove_xml_notatation, split,
    group_by
)


faker = Faker_factory.create()


class Test_keys_to_snake_case( TestCase ):
    def setUp( self ):
        self.d = { 'camelCase': { 'snake_case': { 'PascalCase': '' } } }
        self.expected = {
            'camel_case': { 'snake_case': { 'pascal_case': '' } } }

    def test_should_transform_to_snake_case( self ):
        result = keys_to_snake_case( self.d )
        self.assertEqual( result, self.expected )


class Test_dict(TestCase):
    def setUp( self ):
        pass

    def test_change_keys_dict_simple( self ):
        expected = {
            "new_id": 123
        }
        dict_for_change = {
            "id": 123
        }
        key_for_change = {
            "id": "new_id"
        }
        replace_keys( dict_for_change, key_for_change )
        self.assertEqual( dict_for_change, expected )

    def test_change_keys_list_of_dicts( self ):
        expected = [ {
            "new_id": 123
        } for i in range( 10 ) ]
        dict_for_change = [ {
            "id": 123
        } for i in range( 10 ) ]
        key_for_change = {
            "id": "new_id"
        }
        replace_keys( dict_for_change, key_for_change )
        self.assertEqual( dict_for_change, expected )

    def test_change_keys_dict_with_dicts( self ):
        expected = {
            "new_id": 123,
            "dict": { "bar": "foo" }
        }
        dict_for_change = {
            "id": 123,
            "dict": { "foo": "foo" }
        }
        key_for_change = {
            "id": "new_id",
            "foo": "bar"
        }
        replace_keys( dict_for_change, key_for_change )
        self.assertEqual( dict_for_change, expected )

    def test_change_keys_dict_with_lists( self ):
        expected = {
            "new_id": 123,
            "dict": [ { "bar": "foo" } for i in range( 10 ) ]
        }
        dict_for_change = {
            "id": 123,
            "dict": [ { "foo": "foo" } for i in range( 10 ) ]
        }
        key_for_change = {
            "id": "new_id",
            "foo": "bar"
        }
        replace_keys( dict_for_change, key_for_change )
        self.assertEqual( dict_for_change, expected )

    def test_pop_regex( self ):
        dict_test = {
            "new_id": 123,
            "new_asdf": 234,
            "new_zxcv": 345,
            "new_qwer": 456,
            "stuff": 123
        }
        dict_result_expected = {
            "new_id": 123,
            "new_asdf": 234,
            "new_zxcv": 345,
            "new_qwer": 456,
        }
        dict_expected = {
            "stuff": 123
        }
        result = pop_regex( dict_test, r'new_.*' )
        self.assertEqual( result, dict_result_expected )
        self.assertEqual( dict_test, dict_expected )

    def test_get_regex( self ):
        dict_test = {
            "new_id": 123,
            "new_asdf": 234,
            "new_zxcv": 345,
            "new_qwer": 456,
            "stuff": 123
        }
        dict_result_expected = {
            "new_id": 123,
            "new_asdf": 234,
            "new_zxcv": 345,
            "new_qwer": 456,
        }
        dict_expected = {
            "new_id": 123,
            "new_asdf": 234,
            "new_zxcv": 345,
            "new_qwer": 456,
            "stuff": 123
        }
        result = get_regex( dict_test, r'new_.*' )
        self.assertEqual( result, dict_result_expected )
        self.assertEqual( dict_test, dict_expected )

    def test_rename_keys( self ):
        dict_test = {
            "new_id": 123, "new_asdf": 234, "new_zxcv": 345,
            "new_qwer": 456, "stuff": 123
        }
        dict_result_expected = {
            "id": 123, "asdf": 234, "zxcv": 345,
            "qwer": 456, "stuff": 123
        }
        result = rename_keys(
            dict_test, func=lambda x: x.replace( 'new_', '' ) )
        self.assertEqual( result, dict_result_expected )

    def test_lower_keys( self ):
        dict_test = {
            "ID": 123, "ASDF": 234, "ZXCV": 345,
            "QWER": 456, "STUFF": 123
        }
        dict_result_expected = {
            "id": 123, "asdf": 234, "zxcv": 345,
            "qwer": 456, "stuff": 123
        }
        result = lower_keys( dict_test )
        self.assertEqual( result, dict_result_expected )

    def test_delete_list_of_keys( self ):
        dict_test = {
            "ID": 123, "ASDF": 234, "ZXCV": 345,
            "QWER": 456, "STUFF": 123
        }
        dict_result_expected = {
            "ID": 123, "QWER": 456, "STUFF": 123
        }
        result = delete_list_of_keys( dict_test, 'ASDF', 'ZXCV' )
        self.assertEqual( result, dict_result_expected )

    def test_get_list_of_keys( self ):
        dict_test = {
            "ID": 123, "ASDF": 234, "ZXCV": 345,
            "QWER": 456, "STUFF": 123
        }
        dict_result_expected = { "ASDF": 234, "ZXCV": 345, }
        result = get_list_of_keys( dict_test, 'ASDF', 'ZXCV' )
        self.assertEqual( result, dict_result_expected )

    def test_get_from_dict( self ):
        dict_test = {
            "ID": 123, "ASDF": 234, "ZXCV": 345,
            "QWER": 456, "STUFF": 123
        }
        dict_result_expected = { "a": 234, "b": 345, }
        result = get_from_dict( dict_test, ASDF='a', ZXCV='b' )
        self.assertEqual( result, dict_result_expected )

    def test_remove_value( self ):
        dict_test = {
            "new_id": 123, "new_asdf": 234, "new_zxcv": 345,
            "new_qwer": 456, "stuff": 123
        }
        dict_result_expected = {
            "new_asdf": 234, "new_zxcv": 345, "new_qwer": 456
        }
        result = remove_value( dict_test, 123 )
        self.assertEqual( result, dict_result_expected )

    def test_remove_none( self ):
        dict_test = {
            "new_id": None, "new_asdf": 234, "new_zxcv": 345,
            "new_qwer": 456, "stuff": None
        }
        dict_result_expected = {
            "new_asdf": 234, "new_zxcv": 345, "new_qwer": 456
        }
        result = remove_nones( dict_test )
        self.assertEqual( result, dict_result_expected )

    def test_split( self ):
        dict_test = {
            "new_asdf": 234, "new_zxcv": 345, "new_qwer": 456
        }
        expected = [
            { "new_asdf": 234 }, { "new_zxcv": 345 },
            { "new_qwer": 456 }
        ]
        result = list( split( dict_test ) )
        self.assertEqual( result, expected )

    def test_group_by_with_dict( self ):
        dict_test = { 'id': 1 }
        expected = { 1: dict_test }
        result = group_by( dict_test, 'id' )
        self.assertEqual( result, expected )

    def test_group_by_with_list( self ):
        dict_test = [ { 'id': 1 }, { 'id': 2 }, { 'id': 1 } ]
        expected = { 1: [ dict_test[0], dict_test[2] ], 2: dict_test[1] }
        result = group_by( dict_test, 'id' )
        self.assertEqual( result, expected )


class Test_pipeline( TestCase ):
    def setUp( self ):
        self.example = {
            'nmaprun': {
                '@scanner': 'nmap', '@args': 'nmap',
                '@start': '1556152447',
                '@startstr': 'Wed Apr 24 19:34:07 2019',
                '@version': '7.70',
                '@xmloutputversion': '1.04',
                'verbose': { '@level': '0' },
                'debugging': { '@level': '0' }, }
        }

        self.expected = {
            'nmaprun': {
                'scanner': 'nmap', 'args': 'nmap',
                'start': '1556152447',
                'startstr': 'Wed Apr 24 19:34:07 2019',
                'version': '7.70',
                'xmloutputversion': '1.04',
                'verbose': { 'level': '0' },
                'debugging': { 'level': '0' }, }
        }

    def test_should_remove_xml_garbage( self ):
        result = remove_xml_notatation( self.example )
        self.assertEqual( self.expected, result )



import copy
import math


class End_book( Exception ):
    pass


class Book:
    def __init__(
            self, total_elements, page_size, page=1, min_page=1,
            offset_dict=None ):
        self.total_elements = total_elements
        self.page_size = page_size
        self.page = page
        self.min_page = min_page
        if not offset_dict:
            offset_dict = { 'page': 'page', 'page_size': 'page_size' }
        self.offset_dict = offset_dict

    def __repr__( self ):
        return (
            f"{self.__class__.__name__}( page={self.page}, "
            f"page_size={self.page_size}, "
            f"total_elements={self.total_elements} )"
        )

    def next( self ):
        if self.page >= self.total_pages:
            raise End_book
        self.page += 1

    def prev( self ):
        if self.page <= self.min_page:
            raise End_book
        self.page -= 1

    @property
    def total_pages( self ):
        return math.ceil( self.total_elements / self.page_size )

    @property
    def offset( self ):
        return {
            v: getattr( self, k, None ) for k, v in self.offset_dict.items() }

    def __iter__( self ):
        temp_book = copy.copy( self )
        yield temp_book
        try:
            while True:
                temp_book.next()
                yield temp_book
        except End_book:
            pass



import base64
import binascii
import os
import random
import string

from chibi import b64


def decode( s, code='utf-8' ):
    return s.decode( code )


def generate_string( length=10, letters=string.ascii_letters ):
    """
    Generate a random string

    Arguments
    ---------
    length: int (optional, default 10)
    letters: string (optional, default is all ascii letters)

    Returns
    -------
    string
        random string
    """
    return u''.join(
        random.choice(  letters ) for x in range( length ) )


def generate_email( domain=None, extention=None ):
    """
    Generate random email

    Arguments
    ---------
    domain: string, optional
        Name of domain of email, if no sen is generate randomly
    extention: string, optional
        extencion of the email, if no send is generate randomly

    Returns
    -------
    string
        random string formatted randomly
    """
    if not domain:
        domain = generate_string( 10 )
    if not extention:
        extention = generate_string( 10 )
    return "{name}@{domain}.{extention}".format( **{
        'name': generate_string( 10 ),
        'domain': domain,
        'extention': extention, } )


def generate_password(
        length=20,
        letters=(
            string.ascii_letters + string.digits + string.punctuation ) ):
    """
    Genera un password

    Arguments
    ---------
    length: int
        longitud del password a generar
    letters: string
        cadena con la lista de todas los caracteres que se pueden usar
    """
    return generate_string( length, letters )


def generate_token( length=20 ):
    """
    genera tokens

    Arguments
    ---------
    length: int
        numero de bites usados para generar el token
    """
    return binascii.hexlify( os.urandom( length ) ).decode()


def generate_token_b64( length=8 ):
    """
    genera tokens en base64

    TODO: remplazar los tokens hexadecimales por los de base64

    Arguments
    ---------
    length: int
        numero de bites usados para generar el token
        el default es 8 porque quiero 32 caracteres por default
    """
    return b64.encode( os.urandom( length ) )


def generate_b64_unsecure( length=24 ):
    """
    genera una cadena de base 64 aleatoria no segura

    Arguments
    ---------
    length: int
        numero de bites usados para generar la cadena
    """
    bites = bytearray( random.getrandbits(8) for _ in range( length ) )
    return base64.urlsafe_b64encode( bites ).decode( 'utf-8' )



from unittest import TestCase, skip

from chibi.file.snippets import inflate_dir, join, Chibi_path


@skip( "deprecated" )
class Test_join( TestCase ):

    def setUp( self ):
        self.root = '/'
        self.home = inflate_dir( '~' )

    def test_double_root_is_only_root( self ):
        self.assertEqual( join( self.root, self.root ), join( self.root ) )

    def test_home_and_current_should_end_combined( self ):
        result = join( self.home, 'qwert' )
        self.assertTrue( result.startswith( self.home ) )
        self.assertTrue( result.endswith( 'qwert' ) )

    def test_join_a_chibi_path_and_a_string_should_work( self ):
        a = Chibi_path( '/mnt/hard_drive' )
        b = '12344389.jpg'
        result = join( a, b )
        self.assertEqual( result, '/mnt/hard_drive/12344389.jpg' )



from chibi.file.snippets import ls
from tests.snippet.files import Test_with_files


class Test_ls( Test_with_files ):
    def test_should_list_all_the_files_and_dirs_from_root( self ):
        result = list( ls( self.root_dir ) )
        self.assertTrue(
            len( result ) == ( self.amount_of_dirs + self.amount_of_files ) )

    def test_should_return_a_empty_list( self ):
        result = list( ls( self.empty_folder ) )
        self.assertFalse( result )


class Test_ls_recursive( Test_with_files ):
    def test_should_all_the_folders_and_internal_stuff( self ):
        result = set( ls( self.root_dir, recursive=True ) )

        root = set( ls( self.root_dir ) )
        self.assertFalse( root.difference( result ) )
        for r in root:
            rr = set( ls( r ) )
            self.assertFalse( rr.difference( result ) )

    def test_should_return_a_empty_list( self ):
        result = list( ls( self.empty_folder, recursive=True ) )
        self.assertFalse( result )



from unittest import TestCase

from chibi.file.image import Chibi_image
from tests.snippet.files import Test_with_files


class Test_images( TestCase ):
    def setUp( self ):
        super().setUp()
        self.jpg_path = './tests/file/1529990793499.jpg'
        self.png_path = './tests/file/1535359854403.png'
        self.gif_path = './tests/file/1536637012160.gif'

        self.jpg = Chibi_image( self.jpg_path )
        self.png = Chibi_image( self.png_path )
        self.gif = Chibi_image( self.gif_path )


class Test_eq( Test_images ):
    def test_can_open_the_images( self ):
        self.assertTrue( self.jpg.exists )
        self.assertTrue( self.png.exists )
        self.assertTrue( self.gif.exists )

        self.assertFalse( self.jpg.is_empty )
        self.assertFalse( self.png.is_empty )
        self.assertFalse( self.gif.is_empty )

    def test_should_return_resolution( self ):
        self.assertEqual( self.jpg.dimension, ( 498, 397 ) )
        self.assertEqual( self.png.dimension, ( 418, 498 ) )
        self.assertEqual( self.gif.dimension, ( 728, 720 ) )

    def test_should_get_the_correct_mime( self ):
        self.assertEqual( self.jpg.properties.mime, 'image/jpeg' )
        self.assertEqual( self.png.properties.mime, 'image/png' )
        self.assertEqual( self.gif.properties.mime, 'image/gif' )

    def test_should_get_the_correct_extention( self ):
        self.assertEqual( self.jpg.properties.extension, 'jpg' )
        self.assertEqual( self.png.properties.extension, 'png' )
        self.assertEqual( self.gif.properties.extension, 'gif' )

    def test_should_be_equal( self ):
        self.assertEqual( self.jpg, self.jpg )
        self.assertEqual( self.gif, self.gif )
        self.assertEqual( self.png, self.png )

    def test_should_be_no_equal( self ):
        self.assertNotEqual( self.jpg, self.gif )
        self.assertNotEqual( self.jpg, self.png )

        self.assertNotEqual( self.gif, self.png )
        self.assertNotEqual( self.gif, self.jpg )

        self.assertNotEqual( self.png, self.jpg )
        self.assertNotEqual( self.png, self.gif )


class Test_thumbnails( Test_images, Test_with_files ):
    def test_create_in_diferent_folder( self ):
        thumbnail = self.jpg.thumbnail( self.root_dir )
        self.assertNotEqual( thumbnail.path, self.jpg.path )
        self.assertNotEqual( thumbnail, self.jpg.path )



def string_table_to_dict( s, row_separator='/n', column_separator=' ' ):
    rows = s.split( row_separator )
    headers = rows.pop(0)
    headers = headers.split( column_separator )
    headers = _clean_empty_strings( headers )


def _clean_empty_strings( l ):
    return [ i for i in l if i ]



import logging
import unittest
from tests.snippet.files import Test_with_files
from chibi import config
from chibi.atlas import Chibi_atlas
from chibi.config.config import Logger
from chibi.config import Configuration


class Test_load_config( Test_with_files ):
    def test_should_load_the_settings_from_a_pyton_file( self ):
        python_file = self.root_dir.temp_file( extension='py' )
        self.assertNotIn( 'hello', config.configuration  )
        with python_file as f:
            f.append( 'from chibi.config import configuration\n' )
            f.append( 'configuration.hello = "asdf"' )

        config.load( python_file )
        self.assertEqual( config.configuration.hello, 'asdf' )

    def test_when_read_a_json_should_put_all_the_content_in_the_config( self ):
        json_file = self.root_dir.temp_file( extension='json' )
        self.assertNotIn( 'json_hello', config.configuration  )
        with json_file as f:
            f.write( { 'json_hello': '1234567890' } )

        config.load( json_file )
        self.assertEqual( config.configuration.json_hello, '1234567890' )

    def test_when_read_a_yaml_should_put_all_the_content_in_the_config( self ):
        yaml_file = self.root_dir.temp_file( extension='yaml' )
        self.assertNotIn( 'yaml_hello', config.configuration  )
        with yaml_file as f:
            f.write( { 'yaml_hello': 'qwertyuiop' } )

        config.load( yaml_file )
        self.assertEqual( config.configuration.yaml_hello, 'qwertyuiop' )


class Test_config_default_factory( unittest.TestCase ):
    def setUp( self ):
        super().setUp()
        self.config = Configuration()

    def test_should_create_chibi_atlas_by_default( self ):
        self.assertNotIn( 'new_config', self.config )
        result = self.config.new_config
        self.assertIsInstance( result, Chibi_atlas )


class Test_logger( unittest.TestCase ):
    def setUp( self ):
        super().setUp()
        from chibi.config import configuration
        self.config = configuration
        self.logger = logging.getLogger( 'test.config' )

    def test_should_return_a_instance_of_logger( self ):
        logger = self.config.loggers[ 'test.config' ]
        self.assertIsInstance( logger, Logger )

    def test_should_return_the_current_level( self ):
        logger = self.config.loggers[ 'test.config' ]
        self.assertEqual( logger.level, self.logger.parent.level )

    def test_when_is_set_the_level( self ):
        logger = self.config.loggers[ 'test.config' ]
        current_level = self.logger.level
        logger.level = logging.INFO
        self.assertEqual( logger.level, self.logger.level )
        logger.level = current_level


class Test_envars( unittest.TestCase ):
    def setUp( self ):
        super().setUp()
        from chibi.config import configuration
        self.config = configuration

    def test_envars_should_no_be_empty( self ):
        self.assertTrue( self.config.env_vars )



from unittest import TestCase

from chibi.file import Chibi_path
from chibi.file.snippets import inflate_dir


class Test_inflate_dir( TestCase ):
    def test_inflate_dir_should_inflate_current_dir( self ):
        result = inflate_dir( '.' )
        expected = Chibi_path.current_dir()

        self.assertEqual( result, expected )

    def test_inflate_dir_should_inflate_parent_dir( self ):
        current = inflate_dir( '.' )
        parent = inflate_dir( '..' )

        self.assertTrue( current.startswith( parent ) )

    def test_inflate_dir_should_inflate_home_dir( self ):
        home = inflate_dir( '~' )

        self.assertIn( 'home', home )

    def test_inflate_dir_should_no_change_root( self ):
        root = inflate_dir( '/' )
        self.assertEqual( '/', root )



from unittest import TestCase

from chibi.atlas import Chibi_atlas


class Test_chibi_atlas( TestCase ):
    def setUp( self ):
        self.simple_dict = Chibi_atlas( a=10, b=20, c=30 )

    def test_can_retrieve_by_attr_and_key( self ):
        self.assertEqual( self.simple_dict.a, self.simple_dict[ 'a' ] )
        self.assertEqual( self.simple_dict.b, self.simple_dict[ 'b' ] )
        self.assertEqual( self.simple_dict.c, self.simple_dict[ 'c' ] )

    def test_retrieve_a_not_exists_attribute_raise_AttributeError( self ):
        with self.assertRaises( AttributeError ):
            self.simple_dict.d

    def test_set_an_attribute_dont_exists_should_be_retrieve_by_key( self ):
        new_value = 40
        self.simple_dict.d = new_value
        self.assertEqual( self.simple_dict.d, new_value )
        self.assertEqual( self.simple_dict[ 'd' ], new_value )

    def test_set_a_key_dont_exists_should_be_retrieve_by_attribute( self ):
        new_value = 40
        self.simple_dict[ 'd' ] = new_value
        self.assertEqual( self.simple_dict[ 'd' ], new_value )
        self.assertEqual( self.simple_dict.d, new_value )


class Test_chibi_atlas_deeper( TestCase ):
    def setUp( self ):
        self.d = Chibi_atlas( dict(
            a='a',
            b=dict( b=dict( bb='bb' ) )
        ) )

    def test_the_inner_dicts_should_be_chibi_atlas( self ):
        self.assertIsInstance( self.d.b, Chibi_atlas )
        self.assertIsInstance( self.d.b.b, Chibi_atlas )

    def test_when_assing_a_value_should_be_access_from_root( self ):
        self.d.b.b = dict( a="a" )
        self.assertEqual( self.d.b.b.a, "a" )
        self.assertIsInstance( self.d.b.b, Chibi_atlas )

        self.d.b.b.bb = "another_value"
        self.assertEqual( self.d.b.b.bb, "another_value" )


class Test_chibi_atlas_kw( TestCase ):
    def setUp( self ):
        self.d = Chibi_atlas( **dict(
            a='a',
            b=dict( b=dict( bb='bb' ) )
        ) )

    def test_the_inner_dicts_should_be_chibi_atlas( self ):
        self.assertIsInstance( self.d.b, Chibi_atlas )
        self.assertIsInstance( self.d.b.b, Chibi_atlas )

    def test_when_assing_a_value_should_be_access_from_root( self ):
        self.d.b.b = dict( a="a" )
        self.assertEqual( self.d.b.b.a, "a" )
        self.assertIsInstance( self.d.b.b, Chibi_atlas )

        self.d.b.b.bb = "another_value"
        self.assertEqual( self.d.b.b.bb, "another_value" )


class Test_chibi_atlas_deeper_with_list( TestCase ):
    def setUp( self ):
        self.d = Chibi_atlas( dict(
            a='a',
            b=dict( b=dict( bb='bb' ) )
        ) )
        self.d.l = [
            dict( q=1, w=2 ),
            dict( a=[ dict( z='z', x='x' ) ] )
        ]

    def test_the_inner_dicts_should_be_chibi_atlas( self ):
        self.assertIsInstance( self.d.b, Chibi_atlas )
        self.assertIsInstance( self.d.b.b, Chibi_atlas )
        self.assertIsInstance( self.d.l[0], Chibi_atlas )
        self.assertIsInstance( self.d.l[1], Chibi_atlas )
        self.assertIsInstance( self.d.l[1].a[0], Chibi_atlas )



from chibi.atlas import Chibi_atlas
from chibi.snippet import regex
from chibi.snippet.iter import chunk_each
import re


class Interface( Chibi_atlas ):
    def up( self ):
        raise NotImplementedError

    def down( self ):
        raise NotImplementedError


class Network( Chibi_atlas ):

    @classmethod
    def load_from_string( cls, s ):
        split = s.split( '\n' )
        interfaces_raw = chunk_each(
            split, lambda x: regex.test( r'^\d+:', x  ) )

        result = cls()
        for interface_raw in interfaces_raw:
            interface_raw = "\n".join( interface_raw )
            interface_name = re.search(
                r"^.*: (?P<interface>\w+.+):",
                interface_raw ).groupdict()[ 'interface' ]
            interface = Interface()
            result[ interface_name ] = interface

            interface.ip_v4 = re.search(
                r"inet\s*(?P<ip_v4>\d+.\d+.\d+.\d+/\d+)",
                interface_raw, re.MULTILINE )
            if interface.ip_v4:
                interface.ip_v4 = interface.ip_v4.groupdict()[ 'ip_v4' ]

        return result


class Wireless( Network ):
    pass



from unittest import TestCase

from chibi.snippet.iter import cut_until, chunk_each


class Test_cut_until( TestCase ):
    def setUp( self ):
        self.iterable = 'aaaaabaaabaabab'

    def test_should_return_5a( self ):
        result = cut_until( iter( self.iterable ), lambda x: x != 'a' )
        self.assertEqual( 'aaaaa', "".join( result ) )


class Test_chunk_each( TestCase ):
    def setUp( self ):
        self.iterable = 'baaaabccccbddddbqqqq'

    def test_should_return_5a( self ):
        result = chunk_each( self.iterable, lambda x: x == 'b' )
        for r in result:
            s = "".join( r )
            self.assertEqual( len( s ), 5 )



from unittest import TestCase

from chibi.file.snippets import common_root


class Test_common_root( TestCase ):
    def test_should_return_the_common_root( self ):
        result = common_root( '/usr/var/log', '/usr/var/security' )
        self.assertEqual( '/usr/var/', result  )



from unittest import TestCase

from chibi.snippet.table import _clean_empty_strings


class test_helpers(TestCase):
    def test_generate_string( self ):
        result = _clean_empty_strings( [ 'a', '', 'b', '', 'c', '' ] )
        self.assertListEqual( [ 'a', 'b', 'c' ], result )



import copy
from unittest import TestCase
from chibi.metaphors import Book
from chibi.metaphors.book import End_book


class Test_book( TestCase ):
    def test_total_of_pages( self ):
        book = Book( total_elements=10, page_size=1 )
        self.assertEqual( book.total_pages, 10 )
        book = Book( total_elements=10, page_size=5 )
        self.assertEqual( book.total_pages, 2 )
        book = Book( total_elements=10, page_size=3 )
        self.assertEqual( book.total_pages, 4 )

    def test_next_should_increase_the_page( self ):
        book = Book( total_elements=10, page_size=3 )
        self.assertEqual( book.page, 1 )
        book.next()
        self.assertEqual( book.page, 2 )
        book.next()
        self.assertEqual( book.page, 3 )
        book.next()
        self.assertEqual( book.page, 4 )

    def test_with_next_should_raise_end_of_book( self ):
        book = Book( total_elements=10, page_size=3 )
        self.assertEqual( book.page, 1 )
        book.next()
        self.assertEqual( book.page, 2 )
        book.next()
        self.assertEqual( book.page, 3 )
        book.next()
        self.assertEqual( book.page, 4 )
        with self.assertRaises( End_book ):
            book.next()
        self.assertEqual( book.page, 4 )

    def test_prev_should_decrease_the_page( self ):
        book = Book( total_elements=10, page_size=3, page=4 )
        self.assertEqual( book.page, 4 )
        book.prev()
        self.assertEqual( book.page, 3 )
        book.prev()
        self.assertEqual( book.page, 2 )
        book.prev()
        self.assertEqual( book.page, 1 )

    def test_with_prev_should_raise_end_of_book( self ):
        book = Book( total_elements=10, page_size=3, page=4 )
        self.assertEqual( book.page, 4 )
        book.prev()
        self.assertEqual( book.page, 3 )
        book.prev()
        self.assertEqual( book.page, 2 )
        book.prev()
        self.assertEqual( book.page, 1 )
        with self.assertRaises( End_book ):
            book.prev()
        self.assertEqual( book.page, 1 )

    def test_offset_dict( self ):
        book = Book( total_elements=10, page_size=3 )
        self.assertEqual(
            book.offset,
            { 'page': book.page, 'page_size': book.page_size } )

        book = Book(
            total_elements=10, page_size=3,
            offset_dict={
                'page': 'other_name', 'page_size': 'another_page_size' } )
        self.assertEqual(
            book.offset,
            { 'other_name': book.page, 'another_page_size': book.page_size } )


class Test_copy_book( TestCase ):
    def test_shoudl_copy_the_parameters( self ):
        book = Book( total_elements=10, page_size=3, page=4 )
        copy_book = copy.copy( book )
        self.assertIsNot( book, copy_book )
        self.assertEqual( vars( book ), vars( copy_book ) )

    def test_should_copy_the_offset_dict( self ):
        book = Book(
            total_elements=10, page_size=3,
            offset_dict={
                'page': 'other_name', 'page_size': 'another_page_size' } )

        copy_book = copy.copy( book )
        self.assertIsNot( book, copy_book )
        self.assertEqual( book.offset, copy_book.offset )


class Test_iter_book( TestCase ):
    def setUp( self ):
        self.book = Book( total_elements=10, page_size=3, page=1 )

    def test_should_have_a_iter( self ):
        self.assertTrue( iter( self.book ) )

    def test_should_have_4_pages( self ):
        count = 0
        for page in self.book:
            count += 1

        self.assertEqual( count, 4 )

    def test_should_have_all_pages( self ):
        pages = [ page.page for page in self.book ]
        self.assertEqual( pages, [ 1, 2, 3, 4 ] )

    def test_each_page_generate_should_be_a_intance_of_book( self ):
        for page in self.book:
            self.assertIsInstance( page, Book )

    def test_each_page_generate_should_be_a_diferent_book_to_origin( self ):
        for page in self.book:
            self.assertIsNot( page, self.book )



from chibi.file.other import Chibi_csv, Chibi_yaml
from tests.snippet.files import Test_with_files


class Test_chibi_file_resolution_class( Test_with_files ):
    def test_when_the_file_have_a_extension_yaml_should_be_a_yaml( self ):
        f = self.root_dir.temp_file( extension='yml' )
        file_open = f.open()
        self.assertIsInstance( file_open, Chibi_yaml )
        f = self.root_dir.temp_file( extension='yaml' )
        file_open = f.open()
        self.assertIsInstance( file_open, Chibi_yaml )

    def test_when_the_file_have_a_extension_csv_should_be_a_csv( self ):
        f = self.root_dir.temp_file( extension='csv' )
        file_open = f.open()
        self.assertIsInstance( file_open, Chibi_csv )



import csv
import itertools
from chibi.file import Chibi_file
from chibi.atlas import Chibi_atlas


__all__ = [ 'Chibi_csv' ]


class Chibi_csv( Chibi_file ):
    def __init__( self, *args, has_headers=False, **kw ):
        super().__init__( *args, **kw )
        self.has_headers = has_headers

    def transform_to_xlsx( self, output=None ):
        raise NotImplementedError

    def append( self, data ):
        if isinstance( data, dict ):
            if self.headers:
                row = [ data[ h ] for h in self.headers ]
            else:
                self.append( data.keys() )
                self.append( data )
                return
        else:
            row = data
        with open( self.path, 'a' ) as f:
            writer = csv.writer( f )
            writer.writerow( row )

    def read_as_dict( self ):
        reader = csv.DictReader( self.file )
        for r in reader:
            yield Chibi_atlas( r )
        self.reread()

    def read_as_list( self ):
        reader = csv.reader( self.file )
        for r in reader:
            yield r
        self.reread()

    @property
    def headers( self ):
        reader = csv.reader( self.file )
        try:
            headers = next( reader )
            self.reread()
        except StopIteration:
            return None
        if headers:
            return headers
        return None

    def __getitem__( self, y ):
        reader = csv.reader( self.file )
        try:
            last_row = next( itertools.islice( reader, y, y + 1 ) )
            self.reread()
            return last_row
        except StopIteration:
            raise IndexError( "csv index out of range" )



from tests.snippet.files import Test_with_files
import yaml
from chibi.file import Chibi_file
from chibi.file.other import Chibi_yaml


class Test_chibi_file_yaml( Test_with_files ):
    def setUp( self ):
        super().setUp()
        self.file_path = self.root_dir.temp_file( extension='yml' )
        self.chibi_file = Chibi_file( self.file_path )
        self.data = { 'a': 'a' }
        self.yaml_data = yaml.dump( self.data )

    def test_should_be_the_correct_instance( self ):
        self.assertIsInstance( self.chibi_file, Chibi_yaml )

    def test_should_can_write_and_read_yaml( self ):
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertFalse( whole_file, "el archivo no esta vacio" )
        self.chibi_file.write( self.data )
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertEqual( whole_file, self.yaml_data )

        result = self.chibi_file.read()
        self.assertIsInstance( result, dict )
        self.assertEqual( result, self.data )

    def test_should_can_write_and_read_yaml_safe( self ):
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertFalse( whole_file, "el archivo no esta vacio" )
        self.chibi_file.write( self.data, is_safe=True )
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertEqual( whole_file, self.yaml_data )

        result = self.chibi_file.read()
        self.assertIsInstance( result, dict )
        self.assertEqual( result, self.data )



import datetime

from chibi.atlas import Chibi_atlas, Chibi_atlas_default
from chibi.atlas.tree import Chibi_tree


class Descriptor:
    """
    describe a parameter of class
    """
    default = None

    def __init__( self, name=None, required=False, default=None ):
        self.name = name
        self.required = required
        self.default = default if default is not None else self.default

    def __get__( self, instance, cls ):
        return instance.__dict__.get( self.name, self.default )

    def __set__( self, instance, value ):
        instance.__dict__[ self.name ] = value

    def __repr__( self ):
        return "descriptor.{}".format( type( self ).__name__ )


class Any( Descriptor ):
    def __repr__( self ):
        return "any.{}".format( type( self ).__name__ )


class Kind( Descriptor ):
    """
    descriptor with assert of type
    """
    kind = object

    def __set__( self, instance, value ):
        if not isinstance( value, self.kind ):
            raise TypeError( "Expected {}".format(self.kind ) )
        super().__set__( instance, value )


class Integer( Kind ):
    kind = int
    default = 0


class Float( Kind ):
    kind = float
    default = 0.0


class String( Kind ):
    kind = str
    default = ''


class DateTime( Kind ):
    kind = datetime.datetime
    default = datetime.datetime.now()


class Date( Kind ):
    kind = datetime.date
    default = datetime.datetime.now().date()


class List( list, Descriptor ):
    def __set__( self, instance, value ):
        if value is None:
            value = list()
        super().__set__( instance, list( value ))


class Set( set, Descriptor ):
    def __set__( self, instance, value ):
        if value is None:
            value = set()
        super().__set__( instance, set( value ))


class List_kind_strict( List ):
    def __init__( self, *args, kind=None, **kw ):
        super().__init__( *args, **kw )
        self._kind = kind

    def __set__( self, instance, value ):
        for pos, elem in enumerate( value ):
            if not isinstance( elem, self._kind ):
                raise TypeError(
                    "in the position {}, expected {}".format(
                        pos, self._kind ) )

        super().__set__( instance, value )

    def append( self, item ):
        if self._kind and not isinstance( item, self._kind ):
            raise TypeError("Expected {}".format( self.kind ) )
        super().append( item )


class Dict( Chibi_atlas, Descriptor ):
    def __set__( self, instance, value ):
        if value is None:
            super().__set__( instance, value )
        else:
            super().__set__( instance, Chibi_atlas( value ) )


class Dict_defaults( Chibi_atlas_default, Descriptor ):
    default = None

    def __init__( self, *args, default=None, **kw ):
        super().__init__( *args, **kw )
        if default is None:
            default = self.default
        if not callable( default ):
            default_item = default

            def default_factory():
                return default_item
        else:
            default_factory = default
        self._default_factory = default_factory

    def __set__( self, instance, value ):
        """
        si es None el valor se vuelve a inicializar el dicionario con su
        fabrica por default
        """
        if value is None:
            new_dict = Chibi_atlas_default( self._default_factory )
            super().__set__( instance, new_dict )
        elif not isinstance( value, Chibi_atlas_default ):
            raise TypeError(
                "Expected type of {}".format( Chibi_atlas_default ) )
        else:
            super().__set__( instance, value )


class Tree_simple( Chibi_tree, Descriptor ):
    default = None

    def __init__( self, *args, default_factory=None, **kw ):
        super().__init__( *args, **kw )
        self._default_factory = default_factory

    def __set__( self, instance, value ):
        if value is None:
            new_tree = Chibi_tree( self._default_factory )
            super().__set__( instance, new_tree )
        elif not isinstance( value, Chibi_tree ):
            raise TypeError("Expected type of {}".format( Chibi_tree ) )
        else:
            super().__set__( instance, value )


class Choice( Descriptor ):
    choice = None

    def __init__( self, *args, choice, **kargs ):
        super().__init__( *args, **kargs )
        self.choice = choice

    def __set__( self, instance, value ):
        if value in self.choice:
            super().__set__( instance, value )
        else:
            raise TypeError( (
                "Expected value of ( {} ) but received {}"
            ).format( self.choice, value ) )


class String_choice( Choice, String ):
    pass



from chibi.file.snippets import cd
from chibi.file import Chibi_path
from tests.snippet.files import Test_moving_dir


class Test_cd( Test_moving_dir ):

    def test_ch_can_use_the_simbol_of_home( self ):
        cd( '~' )

    def test_ch_can_use_the_simbol_of_parent_dir( self ):
        cd( '..' )

    def test_ch_for_home_should_change_the_current_directory( self ):
        working_dir = Chibi_path.current_dir()
        cd( '~' )
        new_dir = Chibi_path.current_dir()
        self.assertNotEqual( working_dir, new_dir )

    def test_ch_for_root_should_change_the_current_directory( self ):
        working_dir = Chibi_path.current_dir()
        cd( '/' )
        new_dir = Chibi_path.current_dir()
        self.assertNotEqual( working_dir, new_dir )

    def test_ch_for_parent_should_change_the_current_directory( self ):
        working_dir = Chibi_path.current_dir()
        cd( '..' )
        new_dir = Chibi_path.current_dir()
        self.assertTrue( working_dir.startswith( new_dir ) )

    def test_ch_for_current_dir_should_no_change_the_current_directory( self ):
        working_dir = Chibi_path.current_dir()
        cd( '.' )
        new_dir = Chibi_path.current_dir()
        self.assertTrue( working_dir.startswith( new_dir ) )



from unittest import TestCase

from chibi.pipeline import Pipeline
from chibi.pipeline.pipeline import Pipeline_manager


class Test_pipeline( TestCase ):

    def test_or_create_a_manager( self ):
        p1 = Pipeline()
        p2 = Pipeline()
        m = p1 | p2
        self.assertIsInstance( m, Pipeline_manager )

    def test_or_of_the_class_should_create_a_manager( self ):
        p1 = Pipeline
        p2 = Pipeline
        m = p1 | p2
        self.assertIsInstance( m, Pipeline_manager )


class Test_pipeline_manager( TestCase ):
    def test_should_no_modified_the_childen_of_the_previews_pipelines( self ):
        p1 = Pipeline()
        p2 = Pipeline()
        p3 = Pipeline()
        m1 = p1 | p2
        m2 = p3 | m1

        self.assertEqual( 3, len( m2.children ) )
        self.assertEqual( 2, len( m1.children ) )



from unittest import TestCase
from chibi.snippet import is_type


class Test_is_type(TestCase):
    def setUp( self ):
        pass

    def test_is_iter_list( self ):
        self.assertTrue( is_type.is_iter( [ 1, 2, 3 ] ) )
        self.assertTrue( is_type.is_iter( [] ) )
        self.assertTrue( is_type.is_iter( list() ) )

    def test_is_iter_tuple( self ):
        self.assertTrue( is_type.is_iter( ( 1, 2, 3 ) ) )
        self.assertTrue( is_type.is_iter( tuple() ) )

    def test_is_iter_dict( self ):
        self.assertTrue( is_type.is_iter( { 'foo': 1 } ) )
        self.assertTrue( is_type.is_iter( dict() ) )

    def test_is_iter_string( self ):
        self.assertTrue( is_type.is_iter( "foo" ) )
        self.assertTrue( is_type.is_iter( "" ) )



from unittest import TestCase
from chibi.snippet.func import retry_on_exception, delay


class Test_retry( TestCase ):
    def setUp( self ):
        pass

    def test_should_work( self ):
        @retry_on_exception
        def asdf():
            return "hello my world!!!"
        self.assertTrue( asdf(), 'hello my world!!!' )

    def test_if_not_exception_should_fast_fail( self ):
        exceptions = 0

        @retry_on_exception
        def asdf():
            nonlocal exceptions
            exceptions += 1
            raise Exception( exceptions )

        with self.assertRaises( Exception ):
            asdf()
            self.assertEqual( exceptions, 1 )

    def test_by_default_is_going_to_retry_5_times( self ):
        e = 0

        @retry_on_exception
        def asdf():
            nonlocal e
            e += 1
            raise KeyError( 'asdf' )

        with self.assertRaises( KeyError ):
            asdf()
            self.assertEqual( e, 5 )

    def test_when_is_the_exception_should_do_the_retries( self ):
        es = 0

        @retry_on_exception( exceptions=( KeyError ) )
        def asdf():
            nonlocal es
            es += 1
            raise KeyError( 'asdf' )

        with self.assertRaises( KeyError ):
            asdf()
            self.assertEqual( es, 5 )


class Test_delay( TestCase ):
    def setUp( self ):
        pass

    def test_should_work( self ):
        @delay( seconds=1 )
        def asdf():
            return "hello my world!!!"
        self.assertTrue( asdf(), 'hello my world!!!' )



from chibi.pipeline import Pipeline
from chibi.snippet.dict import remove_xml_notatation


class Remove_xml_garage( Pipeline ):
    def run( self, obj, *args, **kw ):
        return remove_xml_notatation( obj )



import time
import functools


class Max_retries_reach( Exception ):
    pass


def retry_on_exception( f=None, *, exceptions=( Exception, ), max_retries=5 ):
    def _decorate( function, exceptions=exceptions, max_retries=max_retries ):
        @functools.wraps( function )
        def retry_function(
                *args, retries=0, max_retries=max_retries, exception=None,
                exceptions=exceptions,
                **kw ):
            try:
                return function( *args, **kw )
            except Exception as e:
                if exception and not e.__cause__:
                    e.__cause__ = exception
                exception = e

                retries += 1
                if retries <= max_retries:
                    if isinstance( e, exceptions ):
                        return retry_function(
                            *args, retries=retries,
                            max_retries=max_retries, exception=exception,
                            **kw )
                    else:
                        raise exception
                else:
                    raise exception
        return retry_function
    if f:
        return _decorate( f, exceptions=exceptions, max_retries=max_retries )
    return _decorate


def delay( f=None, *, seconds=None ):
    def _decorate( function ):
        @functools.wraps( function )
        def delay_function( *args, **kw ):
            time.sleep( seconds )
            return function( *args, **kw )
        return delay_function
    if f:
        return _decorate( f )
    return _decorate



import grp
import pwd
from unittest import TestCase

from chibi.nix import (
    _parse_passwd, _parse_group, get_passwd, get_group,
    user_exists, group_exists, mem_info
)


class Test_passwd( TestCase ):
    def setUp( self ):
        self.all_passwd = filter(
            lambda x: x.pw_name != "nobody", pwd.getpwall() )

    def test_parse_passwd_should_have_the_same_data_as_the_raw( self ):
        for raw_passwd in self.all_passwd:
            with self.subTest( f"parsing {raw_passwd}" ):
                passwd = _parse_passwd( raw_passwd )

                self.assertEqual( passwd.name, raw_passwd.pw_name )
                self.assertEqual( passwd.passwd, raw_passwd.pw_passwd )
                self.assertEqual( passwd.uid, raw_passwd.pw_uid )
                self.assertEqual( passwd.gid, raw_passwd.pw_gid )
                self.assertEqual( passwd.gecos, raw_passwd.pw_gecos )
                self.assertEqual( passwd.dir, raw_passwd.pw_dir )
                self.assertEqual( passwd.shell, raw_passwd.pw_shell )

    def test_get_passwd_should_raise_a_exeption_if_no_args_are_send( self ):
        with self.assertRaises( ValueError ):
            get_passwd()

    def test_get_passwd_by_uid_should_parse_the_passwd( self ):
        for raw_passwd in self.all_passwd:
            with self.subTest( f"parsing {raw_passwd}" ):
                parse_passwd = _parse_passwd( raw_passwd )
                passwd = get_passwd( uid=raw_passwd.pw_uid )
                self.assertEqual( passwd.keys(), parse_passwd.keys() )

    def test_get_passwd_by_name_should_parse_the_passwd( self ):
        for raw_passwd in self.all_passwd:
            with self.subTest( f"parsing {raw_passwd}" ):
                parse_passwd = _parse_passwd( raw_passwd )
                passwd = get_passwd( name=raw_passwd.pw_name )
                self.assertEqual( passwd.keys(), parse_passwd.keys() )

    def test_get_passwd_by_name_and_uid_should_be_the_same( self ):
        for raw_passwd in self.all_passwd:
            with self.subTest( f"parsing {raw_passwd}" ):
                name_passwd = get_passwd( name=raw_passwd.pw_name )
                uid_passwd = get_passwd( uid=raw_passwd.pw_uid )
                self.assertEqual( name_passwd, uid_passwd )


class Test_group( TestCase ):
    def setUp( self ):
        self.all_group = grp.getgrall()

    def test_parse_group_should_have_the_same_data_as_the_raw( self ):
        for raw_group in self.all_group:
            group = _parse_group( raw_group )

            self.assertEqual( group.name, raw_group.gr_name )
            self.assertEqual( group.passwd, raw_group.gr_passwd )
            self.assertEqual( group.gid, raw_group.gr_gid )
            self.assertEqual( group.mem, raw_group.gr_mem )

    def test_get_group_should_raise_a_exeption_if_no_args_are_send( self ):
        with self.assertRaises( ValueError ):
            get_group()

    def test_get_group_by_gid_should_parse_the_group( self ):
        for raw_group in self.all_group:
            parse_passwd = _parse_group( raw_group )
            group = get_group( gid=raw_group.gr_gid )
            self.assertEqual( group.keys(), parse_passwd.keys() )

    def test_get_group_by_name_should_parse_the_group( self ):
        for raw_group in self.all_group:
            parse_passwd = _parse_group( raw_group )
            group = get_group( name=raw_group.gr_name )
            self.assertEqual( group.keys(), parse_passwd.keys() )

    def test_get_group_by_name_and_gid_should_be_the_same( self ):
        for raw_group in self.all_group:
            name_passwd = get_group( name=raw_group.gr_name )
            gid_passwd = get_group( gid=raw_group.gr_gid )
            self.assertEqual( name_passwd, gid_passwd )


class Test_exists( TestCase ):
    def test_user_exists_only_name( self ):
        self.assertTrue( user_exists( name='root' ) )

    def test_user_exists_only_uid( self ):
        self.assertTrue( user_exists( uid=0 ) )

    def test_user_exists_name_and_uid( self ):
        self.assertTrue( user_exists( uid=0, name='root' ) )

    def test_user_exists_name_and_uid_if_one_no_exits_should_be_false( self ):
        self.assertFalse( user_exists( uid=1, name='root' ) )
        self.assertFalse( user_exists( uid=0, name='toor' ) )

    def test_group_exists_only_name( self ):
        self.assertTrue( group_exists( name='root' ) )

    def test_group_exists_only_uid( self ):
        self.assertTrue( group_exists( gid=0 ) )

    def test_group_exists_name_and_uid( self ):
        self.assertTrue( group_exists( gid=0, name='root' ) )

    def test_group_exists_name_and_uid_if_one_no_exits_should_be_false( self ):
        self.assertFalse( group_exists( gid=1, name='root' ) )
        self.assertFalse( group_exists( gid=0, name='toor' ) )


class Test_mem_info( TestCase ):
    def test_mem_info( self ):
        result = mem_info()
        self.assertIn( 'mem_total', result )
        self.assertGreater( result.mem_total, 1000 )



import re


camelcase = re.compile( r'([A-Z][a-z]+)|([a-z]+[A-Z])' )
separate_camelcase_phase_1 = re.compile( r'(.)([A-Z][a-z]+)' )
separate_camelcase_phase_2 = re.compile( r'([a-z0-9])([A-Z])' )

regex_email_validate = r'''^(?:(?:[\w`~!#$%^&*\-=+;:{}'|,?\/]+(?:(?:\.(?:"(?:\\?[\w`~!#$%^&*\-=+;:{}'|,?\/\.()<>\[\] @]|\\"|\\\\)*"|[\w`~!#$%^&*\-=+;:{}'|,?\/]+))*\.[\w`~!#$%^&*\-=+;:{}'|,?\/]+)?)|(?:"(?:\\?[\w`~!#$%^&*\-=+;:{}'|,?\/\.()<>\[\] @]|\\"|\\\\)+"))@(?:[a-zA-Z\d\-]+(?:\.[a-zA-Z\d\-]+)*|\[\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\])$'''  # noqa

email_validator = re.compile( regex_email_validate )


def test( regex, string ):
    """
    check if the string is part of the language of regex

    Arguemtens
    ----------
    regex: string or regex
    string: string

    Returns
    -------
    Bool
    """
    if isinstance( regex, str ):
        regex = re.compile( regex )
    return regex.match( string ) is not None


def is_email( email ):
    """
    check if the email is valid

    Arguments
    ---------
    email: string

    Returns
    -------
    Bool
    """
    return test( email_validator, email )


def is_camelcase( text ):
    return test( camelcase, text )



from .config import (
    __all__ as config_all,
    Configuration, Logger_configuration, Env_vars )
from .logger import *  # noqa: F403
from chibi.file import Chibi_path


__all__ = config_all + logger.__all__  # noqa: F405


configuration = Configuration(
    loggers=Logger_configuration(),
    env_vars=Env_vars(),
)


def default_file_load():
    config_home = configuration.env_vars.HOME
    if not config_home:
        return

    config_home = Chibi_path( configuration.env_vars.XDG_CONFIG_HOME )
    if not config_home:
        config_home = Chibi_path( '~/.config' )

    config_home += 'chibi'
    config_file = config_home + 'chibi.py'
    if not config_home.exists:
        return
        # config_home.mkdir()

    if not config_file.exists:
        return
        # config_file.touch()
    load( config_file )


def load( path ):
    configuration.load( path )


default_file_load()



from faker import Factory as Faker_factory

from chibi.file.path import Chibi_path
from chibi.file.snippets import exists
from tests.snippet.files import Test_with_files


faker = Faker_factory.create()


class Test_glob( Test_with_files ):
    def test_ls_should_return_the_path( self ):
        all_tmps = self.root_dir + 'tmp*'
        ls = all_tmps.ls()
        for l in ls:
            self.assertNotIn( '*', l )

    def test_expand_should_have_all_files( self ):
        source = Chibi_path( self.folder_with_files_with_content ) + '*'
        result = set( source.expand )
        self.assertEqual(
            result, set( self.folder_with_files_with_content.ls() ) )

    def test_ls_with_glob_on_start_should_have_all_items( self ):
        source = Chibi_path( self.folder_with_files_with_content ) + '*'
        result = set( source.ls() )
        self.assertEqual(
            result, set( self.folder_with_files_with_content.ls() ) )

    def test_copy( self ):
        dest = Chibi_path( self.root_dir ) + 'hola'
        self.assertFalse( exists( dest ) )
        source = Chibi_path( self.folder_with_files_with_content ) + '*'
        source.copy( dest )
        self.assertEqual(
            set( f.base_name for f in source.ls() ),
            set( f.base_name for f in dest.ls() ) )

    def test_move( self ):
        dest = Chibi_path( self.root_dir ) + 'hola'
        self.assertFalse( exists( dest ) )
        source = Chibi_path( self.folder_with_files_with_content ) + '*'
        files = set( f.base_name for f in source.ls() )
        source.move( dest )
        self.assertEqual(
            files, set( f.base_name for f in dest.ls() ) )
        self.assertFalse( set( f.base_name for f in source.ls() ) )



from chibi.file.snippets import cd
from chibi.file import Chibi_path
from tests.snippet.files import Test_moving_dir


class Test_current_dir( Test_moving_dir ):
    def move_root( self ):
        cd( '/' )

    def move_home( self ):
        cd( '~' )

    def test_current_dir_should_return_a_string( self ):
        result = Chibi_path.current_dir()
        self.assertIsInstance( result, str )

    def test_current_dir_should_no_return_empty( self ):
        result = Chibi_path.current_dir()
        self.assertTrue( result )

    def test_current_dir_is_different_when_move_to_home( self ):
        first_current_dir = Chibi_path.current_dir()
        self.move_home()
        home_dir = Chibi_path.current_dir()
        self.assertNotEqual( first_current_dir, home_dir )

    def test_current_dir_is_different_when_move_to_root( self ):
        first_current_dir = Chibi_path.current_dir()
        self.move_root()
        root_dir = Chibi_path.current_dir()
        self.assertNotEqual( first_current_dir, root_dir )

    def test_should_no_be_root( self ):
        directory = Chibi_path.current_dir()
        self.assertNotEqual( directory, '/' )



from unittest import TestCase

from chibi.snippet.string import (
    replace_with_dict, get_the_number_of_parameters, split_in_table,
    remove_inner_space
)


class Test_sp_random( TestCase ):
    def setUp( self ):
        pass

    def test_generate_string( self ):
        string_test = 'case_1, asdf, case_2'
        string_espected = 'case_one, asdf, case_two'
        dict_test = {
            'case_1': 'case_one',
            'case_2': 'case_two',
        }

        result = replace_with_dict( string_test, dict_test )
        self.assertEqual( result, string_espected )

    def test_the_number_parameters( self ):
        strings = [
            "{}",
            "pi-piru {}",
            "{} pi-piru {}",
        ]
        expected = [ 1, 1, 2 ]
        for s, e in zip( strings, expected ):
            r = get_the_number_of_parameters( s )
            self.assertEqual(
                r, e,
                "\nfallo con la cadena {} regreso {} se esperaban {}".format(
                    s, r, e ) )


class Test_split_table( TestCase ):
    def setUp( self ):
        self.table_1 = (
            "Pipiru:     piru-piru        pipiru-pii\n"
            "Pipiru:   piru-piru pipiru-pii\n" )
        self.expected_1 = [
            [ 'Pipiru:', 'piru-piru', 'pipiru-pii' ],
            [ 'Pipiru:', 'piru-piru', 'pipiru-pii' ], [],
        ]

    def test_table_1( self ):
        result = split_in_table( self.table_1 )
        self.assertEqual( self.expected_1, result )


class Test_remove_inner_space( TestCase ):
    def test_should_remove_inner_spaces( self ):
        self.assertEqual( "asdf asdf", remove_inner_space( "asdf   asdf" ) )

    def test_should_remove_inner_new_lines( self ):
        self.assertEqual( "asdf asdf", remove_inner_space( "asdf \n  asdf" ) )



from unittest import TestCase
from chibi import madness


class Test_generate_token_b64( TestCase ):
    def test_should_no_return_a_empty_string( self ):
        self.assertTrue( madness.string.generate_token_b64 )


class Test_generate_b64_unsecure( TestCase ):
    def test_should_no_return_a_empty_string( self ):
        self.assertTrue( madness.string.generate_b64_unsecure )


class Test_generate_string( TestCase ):
    def test_should_no_return_a_empty_string( self ):
        self.assertTrue( madness.string.generate_string )


class Test_generate_email( TestCase ):
    def test_should_no_return_a_empty_string( self ):
        self.assertTrue( madness.string.generate_email )


class Test_generate_password( TestCase ):
    def test_should_no_return_a_empty_string( self ):
        self.assertTrue( madness.string.generate_password )



from unittest import TestCase
from chibi.atlas import Chibi_atlas_ignore_case


class Test_ignore_case:
    def test_can_retrive_upperkeys_and_are_equal_to_lower( self ):
        self.assertEqual( self.simple_dict.A, self.simple_dict.a )
        self.assertEqual( self.simple_dict.B, self.simple_dict.b )
        self.assertEqual( self.simple_dict.C, self.simple_dict.c )

    def test_can_retrieve_by_attr_and_key( self ):
        self.assertEqual( self.simple_dict[ 'A' ], self.simple_dict[ 'a' ] )
        self.assertEqual( self.simple_dict[ 'B' ], self.simple_dict[ 'b' ] )
        self.assertEqual( self.simple_dict[ 'C' ], self.simple_dict[ 'c' ] )

    def test_set_an_attribute_in_upper_case_can_should_be_retrive_in_lower(
            self ):

        new_value = 40
        self.simple_dict.D = new_value
        self.assertEqual( self.simple_dict.d, new_value )
        self.assertEqual( self.simple_dict[ 'd' ], new_value )

    def test_set_a_key_in_upper_case_should_can_be_retrieve_in_lower_case(
            self ):

        new_value = 40
        self.simple_dict[ 'D' ] = new_value
        self.assertEqual( self.simple_dict[ 'd' ], new_value )
        self.assertEqual( self.simple_dict.d, new_value )


class Test_chibi_atlas_ignore_case( TestCase, Test_ignore_case ):
    def setUp( self ):
        self.simple_dict = Chibi_atlas_ignore_case( A=10, B=20, C=30 )


class Test_chibi_atlas_first_parameters_is_a_dict_work(
        TestCase, Test_ignore_case ):
    def setUp( self ):
        self.simple_dict = Chibi_atlas_ignore_case( dict( A=10, B=20, C=30 ) )



def is_iter( obj ):
    """
    evalua si el objeto se puede iterar

    Arguments
    ---------
    obj: any type

    Returns
    -------
    bool
        True if the object is iterable
    """
    try:
        iter( obj )
        return True
    except TypeError:
        return False


def is_number( obj ):
    """
    evalua si es objecto es un numero

    Arguments
    ---------
    obj: any type

    Returns
    -------
    bool
        True if the object is a number
    """
    return isinstance( obj, ( int, float ) )



import unittest
from chibi.atlas.multi import Chibi_atlas_multi


class Test_chibi_atlas_multi( unittest.TestCase ):
    def test_should_work_normal( self ):
        m = Chibi_atlas_multi()
        m[ 'a' ] = 'a'
        m[ 'b' ] = 'b'
        m[ 'c' ] = 'c'

        self.assertEqual( { 'a': 'a', 'b': 'b', 'c': 'c' }, m )

    def test_should_append_the_repeated_keys( self ):
        m = Chibi_atlas_multi()
        m[ 'a' ] = 'a'
        m[ 'a' ] = 'b'
        m[ 'a' ] = 'c'

        self.assertEqual( { 'a': [ 'a', 'b', 'c', ] }, m )

    def test_when_update_should_add_a_list( self ):
        m = Chibi_atlas_multi()
        m[ 'a' ] = 'a'
        d = { 'a': 'b' }
        m.update( d )
        self.assertEqual( m, { 'a': [ 'a', 'b' ] } )



import os
import random
from unittest import TestCase

from faker import Factory as Faker_factory

from chibi.file import Chibi_file
from chibi.file.path import Chibi_path
from chibi.file.snippets import exists, ls, file_dir
from tests.snippet.files import Test_with_files


class Dump_chibi_file( Chibi_file ):
    pass


faker = Faker_factory.create()


class Test_path( Test_with_files ):
    def setUp( self ):
        super().setUp()
        self.path = Chibi_path( self.root_dir )

    def test_can_add_path_and_str( self ):
        dirs = ls( self.root_dir )
        for d in dirs:
            result = self.path + d
            self.assertEqual(
                result, os.path.join( str( self.path ), str( d ) ) )

    def test_ls_work( self ):
        dirs = list( self.path.ls() )
        self.assertGreaterEqual( len( dirs ), 1 )

    def test_ls_return_chibi_path( self ):
        dirs = list( self.path.ls() )
        self.assertGreaterEqual( len( dirs ), 1 )
        for d in dirs:
            self.assertIsInstance( d, Chibi_path )

    def test_find_work( self ):
        result = list( self.path.find() )
        for l in self.path.ls():
            self.assertIn( l, result )

    def test_mkdir_should_create_the_folder( self ):
        new_dir = Chibi_path( self.dirs[0] ) + 'asdf'
        if exists( new_dir ):
            self.fail(
                "el directiorio {} ya existe usar"
                "otro nombre".format( new_dir ) )
        new_dir.mkdir()
        if not exists( new_dir ):
            self.fail(
                "no se creo el directorio {} ".format( new_dir ) )

    def test_copy_folder_should_copy_all_folder( self ):
        dest = Chibi_path( self.root_dir ) + 'hola'
        self.assertFalse( exists( dest ) )
        source = Chibi_path( self.folder_with_files_with_content )
        source.copy( dest )
        self.assertTrue( exists( dest ) )

        self.assertEqual(
            len( set( ls( source ) ) ), len( set( ls( dest ) ) ) )

    def test_copy_folder_on_exist_folder_should_copy_the_files( self ):
        dest = Chibi_path( self.root_dir ) + 'hola'
        self.assertFalse( exists( dest ) )
        source = Chibi_path( self.folder_with_files_with_content )
        source.copy( dest )
        self.assertEqual(
            set( f.base_name for f in source.ls() ),
            set( f.base_name for f in dest.ls() ) )
        for d in ( dest + '*' ).ls():
            d.delete()

        items = list( dest.ls() )
        self.assertFalse( items )
        source.copy( dest )
        self.assertEqual(
            set( f.base_name for f in source.ls() ),
            set( f.base_name for f in dest.ls() ) )

    def test_copy_to_a_existen_dir_should_override_the_current_files( self ):
        dest = Chibi_path( self.root_dir ) + 'hola'
        source = Chibi_path( self.folder_with_files_with_content )
        source.copy( dest )
        source.copy( dest )

    def test_copy_file_should_copy_the_file( self ):
        source = Chibi_path( random.choice( self.files_with_content ) )
        dest = Chibi_path( self.root_dir ) + faker.file_name()
        self.assertFalse( exists( dest ) )

        source.copy( dest, verbose=False )
        self.assertTrue( exists( dest ) )
        s = Chibi_file( source )
        d = Chibi_file( dest )
        self.assertEqual( s.file.read(), d.file.read() )

    def test_when_delete_a_file_should_no_exists( self ):
        path = Chibi_path( random.choice( self.files ) )
        self.assertTrue( exists( path ) )
        path.delete()
        self.assertFalse( exists( path ) )

    def test_whe_delete_a_dir_should_removed( self ):
        path = Chibi_path( random.choice( self.dirs ) )
        self.assertTrue( exists( path ) )
        path.delete()
        self.assertFalse( exists( path ) )


class Test_path_with_files( Test_with_files ):
    def test_if_path_is_a_file_should_only_use_the_dir( self ):
        for f in self.files:
            d = file_dir( f )
            p_f = Chibi_path( f )
            self.assertEqual( p_f + "another", os.path.join( d, 'another' ) )


class Test_path_relative( TestCase ):
    def test_can_add_path_and_str( self ):
        path = Chibi_path( '/usr/var/log' )
        result = path.relative_to( '/usr/var' )
        self.assertEqual( 'log', result )


class Test_path_chown( Test_with_files ):
    def test_verbose_when_no_change_the_owners( self ):
        f = Chibi_path( self.files[0] )
        current_stat = f.properties
        with self.assertLogs( level='INFO' ) as cm:
            f.chown()
        output = cm.output[0]

        self.assertIn( 'permanece', output )
        self.assertIn(
            '{}:{}'.format(
                current_stat.user.name, current_stat.group.name ),
            output )


class Test_path_chmod( Test_with_files ):
    def test_verbose_when_no_change_the_owners( self ):
        f = Chibi_path( self.files[0] )
        current_stat = f.properties

        f.chmod( 0o755 )
        new_stat = f.properties
        self.assertNotEqual( current_stat.mode, new_stat.mode )


class Test_path_extension( Test_with_files ):
    def test_should_replace_the_extension( self ):
        f = Chibi_path( self.files[0] )
        self.assertFalse( f.endswith( '.ext' ) )
        f = f.replace_extensions( 'ext' )
        self.assertTrue( f.endswith( '.ext' ) )

    def test_should_add_the_extension( self ):
        f = Chibi_path( self.files[0] )
        self.assertFalse( f.endswith( '.ext' ) )
        f = f.add_extensions( 'ext' )
        self.assertTrue( f.endswith( '.ext' ) )


class Test_path_file_name( Test_with_files ):
    def test_should_only_return_the_file_name( self ):
        f = Chibi_path( self.files[0] )
        self.assertFalse( f.endswith( '.ext' ) )
        expected = f.base_name
        f = f.replace_extensions( 'ext' )
        result = f.file_name
        self.assertEqual( expected, result )


class Test_move( Test_with_files ):
    def test_when_move_a_empty_file_should_create_a_new_empty_file( self ):
        file = Chibi_path( random.choice( self.files ) )
        dest = Chibi_path( self.root_dir ) + faker.file_name()
        self.assertFalse( dest.exists )

        file.move( dest )

        self.assertFalse( file.exists )
        with open( str( dest ) ) as file_dest:
            self.assertFalse( file_dest.read() )

    def test_move_file_to_folder( self ):
        file = Chibi_path( random.choice( self.files ) )
        dest = Chibi_path( random.choice( self.dirs ) )

        file.move( dest )

        self.assertFalse( file.exists )
        self.assertTrue( dest.exists )

    def test_move_folder_to_another_another_name( self ):
        folder = Chibi_path( random.choice( self.dirs ) )
        dest = Chibi_path( self.root_dir ) + faker.name()

        folder.move( dest )

        self.assertFalse( folder.exists )
        self.assertTrue( dest.exists )


class Test_contains( Test_with_files ):
    def test_child_path_parent_path_should_be_true( self ):
        child = Chibi_path( random.choice( self.files ) )
        parent = Chibi_path( self.root_dir )
        self.assertIn( child, parent )

    def test_parent_in_child_should_be_false( self ):
        child = Chibi_path( random.choice( self.files ) )
        parent = Chibi_path( self.root_dir )
        self.assertNotIn( parent, child )


class Test_context_manager( Test_with_files ):
    def test_when_enter_should_return_a_chibi_file( self ):
        with self.root_dir.temp_file() as f:
            self.assertIsInstance( f, Chibi_file )

    def test_when_enter_raise_a_exception_should_do_nothing( self ):
        with self.assertRaises( ZeroDivisionError ):
            with self.root_dir.temp_file():
                0 / 0


class Test_made_safe( Test_with_files ):
    def test_should_clean_the_characters( self ):
        path = Chibi_path( 'asdf*asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf<asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf>asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf:asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf"asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf|asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf?asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )
        path = Chibi_path( 'asdf?*<>asdf' ).made_safe()
        self.assertEqual( 'asdfasdf', path )


class Test_path_open( Test_with_files ):
    def test_when_open_a_file_should_retunr_a_chibi_file( self ):
        f = self.files[0]
        self.assertIsInstance( f.open(), Chibi_file )

    def test_when_pass_the_class_should_return_the_expected_class( self ):
        f = self.files[0]
        self.assertIsInstance(
            f.open( chibi_file_class=Dump_chibi_file ), Dump_chibi_file )

    def test_the_chibi_path_can_carrie_the_chibi_file_whant_to_be_used( self ):
        f = Chibi_path( self.files[0], chibi_file_class=Dump_chibi_file )
        self.assertIsInstance( f.open(), Dump_chibi_file )



from unittest import TestCase

from chibi.pipeline.xml import Remove_xml_garage


example = {
    'nmaprun': {
        '@scanner': 'nmap', '@args': 'nmap',
        '@start': '1556152447',
        '@startstr': 'Wed Apr 24 19:34:07 2019',
        '@version': '7.70',
        '@xmloutputversion': '1.04',
        'verbose': { '@level': '0' },
        'debugging': { '@level': '0' }, }
}

expected = {
    'nmaprun': {
        'scanner': 'nmap', 'args': 'nmap',
        'start': '1556152447',
        'startstr': 'Wed Apr 24 19:34:07 2019',
        'version': '7.70',
        'xmloutputversion': '1.04',
        'verbose': { 'level': '0' },
        'debugging': { 'level': '0' }, }
}


class Test_pipeline( TestCase ):
    def test_should_remove_xml_garbage( self ):
        result = Remove_xml_garage().run( example )
        self.assertEqual( result, expected )



import re

from chibi.atlas import Chibi_atlas
from chibi.atlas.multi import Chibi_atlas_multi
from chibi.file import Chibi_file
from chibi.snippet.iter import chunk_each


__all__ = [ 'Chibi_systemd' ]


category_regex = re.compile( r"\[.*\]" )


class Chibi_systemd( Chibi_file ):
    def read( self ):
        data = super().read()
        result = Chibi_atlas()
        lines = filter( bool, data.split( '\n' ) )
        lines = filter( lambda x: not x.startswith( '#' ), lines )
        chunks = chunk_each( lines, lambda x: category_regex.match( x ) )
        for section in chunks:
            section_key = section[0][1:-1].strip()
            section_data = section[1:]
            section_dict = Chibi_atlas_multi()
            for line in section_data:
                key, value = line.split( '=', 1 )
                key = key.strip()
                value = value.strip()
                section_dict[key] = value
            result[ section_key.lower() ] = section_dict
        return result

    def write( self, data ):
        result = ''
        sort_order = [ 'unit', 'service' ]
        other_keys = ( k for k in data.keys() if k not in sort_order )
        for k in sort_order:
            section = data[ k ]
            section_str = self._transform_section( k, section )
            result += section_str
        for k in other_keys:
            section = data[ k ]
            section_str = self._transform_section( k, section )
            result += section_str
        super().write( result )

    def _transform_section( self, name, section ):
        name = name.capitalize()
        result = f"[{name}]\n"
        for k, v in section.items():
            if isinstance( v, list ):
                for inner in v:
                    result += f"{k}={inner}\n"
            else:
                result += f"{k}={v}\n"
        return result



import logging


__all__ = [ 'basic_config', 'silent' ]


def basic_config( level=logging.INFO ):
    from . import configuration
    if configuration.env_vars.PYTHON_UNITTEST_LOGGER:
        level = configuration.env_vars.PYTHON_UNITTEST_LOGGER
    logger_formarter = '%(levelname)s %(asctime)s %(name)s %(message)s'
    logging.basicConfig( level=level, format=logger_formarter )


def silent():
    basic_config( logging.ERROR )



import base64


def encode( s ):
    if isinstance( s, str ):
        s = s.encode( 'utf-8' )
    return base64.urlsafe_b64encode( s ).decode( 'utf-8' )


def decode( s ):
    return base64.urlsafe_b64decode( s ).decode( 'utf-8' )



from unittest import TestCase

from chibi.snippet import regex


class test_regex(TestCase):

    def setUp(self):
        self.valid_email = [
            '"abcdefghixyz"@example1.com', 'abc."defghi".xyz@example-1.com',
            'niceandsimple@example.com', 'very.common@example.com',
            'a.little.lengthy.but.fine@dept.example.com',
            'disposable.style.email.with+symbol@example.com',
            'other.email-with-dash@example.com',
            '"much.more unusual"@example.com',
            '"very.unusual.@.unusual.com"@example.com',
            ('"very.(),:;<>[]\\".VERY.\\"very@\\ \\"very\\".unusual"'
             '@strange.example.com' ),
            'admin@mailserver1', '#!$%&\'*+-/=?^_`{}|~@example.org',
            '"()<>[]:,;@\\\"!#$%&\'*+-/=?^_`{}| ~.a"@example.org',
            '" "@example.org', 'jsmith@example.org', 'jsmith@[192.168.2.1]',
        ]

        self.invalid_email = [
            'abc\"def\"ghi@example.com', 'abc"defghi"xyz@example.com',
            'Abc.example.com', 'A@b@c@example.com',
            'a"b(c)d,e:f;g<h>i[j\\k]l@example.com',
            'just"not"right@example.com',
            'this is"not\allowed@example.com',
            'this\\ still\"not\\allowed@example.com',
            'john..doe@example.com', 'john.doe@example..com', ]

    def test_test(self):
        re = r'10*1'
        inner = 0
        s = '1%s1'

        for i in range(10):
            regex.test(re, s % (inner * i))

    def test_email_validate(self):
        for email in self.valid_email:
            self.assertTrue(
                regex.is_email( email ),
                "this is a valid email: {}".format( email ) )

    def test_email_validate_with_invalid(self):
        for email in self.invalid_email:
            self.assertFalse(
                regex.is_email( email ),
                "this is a invalid email: {}".format( email ) )



from unittest import TestCase
from chibi.file.temp import Chibi_temp_path
from chibi.file.other import Chibi_systemd


content = """
[Unit]
Description=Network Manager
Documentation=man:NetworkManager(8)
Wants=network.target
After=network-pre.target dbus.service
Before=network.target

[Service]
Type=dbus
BusName=org.freedesktop.NetworkManager
ExecReload=/usr/bin/busctl call org.freedesktop.NetworkManager
#ExecReload=/bin/kill -HUP $MAINPID
ExecStart=/usr/bin/NetworkManager --no-daemon
Restart=on-failure
# NM doesn't want systemd to kill its children for it
KillMode=process
CapabilityBoundingSet=CAP_NET_ADMIN CAP_DAC_OVERRIDE CAP_NET_RAW

ProtectSystem=true
ProtectHome=read-only

[Install]
WantedBy=multi-user.target
Also=NetworkManager-dispatcher.service

# We want to enable NetworkManager-wait-online.service whenever this service
# is enabled. NetworkManager-wait-online.service has
# WantedBy=network-online.target, so enabling it only has an effect if
# network-online.target itself is enabled or pulled in by some other unit.
Also=NetworkManager-wait-online.service
"""


class Test_chibi_service( TestCase ):
    def setUp( self ):
        self.folder = Chibi_temp_path()
        self.file_service = self.folder.temp_file( extension='service' )
        with open( self.file_service, 'w' ) as f:
            f.write( content )

    def test_should_be_a_dict( self ):
        service = Chibi_systemd( self.file_service )
        result = service.read()
        self.assertIsInstance( result, dict )

    def test_should_have_the_3_sections( self ):
        service = Chibi_systemd( self.file_service )
        result = service.read()
        self.assertIn( 'unit', result )
        self.assertIn( 'service', result )
        self.assertIn( 'install', result )

    def test_unit_should_have_the_expected_keys( self ):
        service = Chibi_systemd( self.file_service )
        result = service.read()
        expected = dict(
            Description='Network Manager',
            Documentation='man:NetworkManager(8)',
            Wants='network.target',
            After='network-pre.target dbus.service',
            Before='network.target' )
        self.assertIn( 'unit', result )
        self.assertEqual( result.unit, expected )

    def test_write_should_write_update_the_file_using_systemd_format( self ):
        service = Chibi_systemd( self.file_service )
        result = service.read()
        result.service.TimeoutSec = '900'
        service.write( result )
        result_after_save = service.read()
        self.assertEqual( result, result_after_save )



from unittest import TestCase

from chibi.net.network.interface import Network


class Test_load_interface_from_ip( TestCase ):
    def setUp( self ):
        self.example = (
            '1: lo: <LOOPBACK,UP,LOWER_UP> mtu 65536 qdisc noqueue '
            'state UNKNOWN group default qlen 1000\n    '
            'link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00\n    '
            'inet 127.0.0.1/8 scope host lo\n       '
            'valid_lft forever preferred_lft forever\n    '
            'inet6 ::1/128 scope host \n       '
            'valid_lft forever preferred_lft forever\n2: enp0s25: '
            '<NO-CARRIER,BROADCAST,MULTICAST,UP> mtu 1500 qdisc fq_codel '
            'state DOWN group default qlen 1000\n    '
            'link/ether 00:21:cc:65:7a:0a brd ff:ff:ff:ff:ff:ff\n3: '
            'wlp3s0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc mq '
            'state UP group default qlen 1000\n    '
            'link/ether 74:e5:0b:03:07:30 brd ff:ff:ff:ff:ff:ff\n   '
            ' inet 10.106.70.59/21 brd 10.106.71.255 scope global dynamic '
            'noprefixroute wlp3s0\n       valid_lft 1889sec preferred_lft '
            '1889sec\n    inet6 fe80::727a:540:5d5e:2b55/64 scope '
            'link noprefixroute \n       valid_lft forever preferred_lft '
            'forever\n4: vboxnet0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500'
            'qdisc fq_codel state UP group default qlen 1000\n    '
            'link/ether 0a:00:27:00:00:00 brd ff:ff:ff:ff:ff:ff\n    '
            'inet 192.168.56.1/24 brd 192.168.56.255 scope global vboxnet0'
            '\n       valid_lft forever preferred_lft forever\n    '
            'inet6 fe80::800:27ff:fe00:0/64 scope link \n       '
            'valid_lft forever preferred_lft forever\n5: vboxnet1:'
            ' <BROADCAST,MULTICAST> mtu 1500 qdisc noop state DOWN '
            'group default qlen 1000\n    '
            'link/ether 0a:00:27:00:00:01 brd ff:ff:ff:ff:ff:ff\n'
        )

    def test_should_return_4_interfaces( self ):
        result = Network.load_from_string( self.example )
        self.assertIn( 'lo', result )
        self.assertIn( 'vboxnet1', result )
        self.assertIn( 'wlp3s0', result )
        self.assertIn( 'enp0s25', result )
        self.assertIn( 'vboxnet1', result )

    def test_the_ip_should_be_the_expected( self ):
        result = Network.load_from_string( self.example )
        self.assertEqual( result.lo.ip_v4, '127.0.0.1/8' )
        self.assertEqual( result.enp0s25.ip_v4, None )
        self.assertEqual( result.wlp3s0.ip_v4, '10.106.70.59/21' )
        self.assertEqual( result.vboxnet0.ip_v4, '192.168.56.1/24' )
        self.assertEqual( result.vboxnet1.ip_v4, None )



from chibi.file.snippets import (
    ls, find, base_name, is_a_file, is_a_folder, find_only_files,
    find_only_folders,
)
from tests.snippet.files import Test_with_files


class Test_find( Test_with_files ):
    defined_folders = [ 'a', 'b' ]
    defined_files = [ 'w', 'q', 'a/e', 'b/r' ]

    def test_without_search_should_be_the_same_to_ls( self ):
        result = set( find( self.root_dir ) )
        ls_result = set( ls( self.root_dir, recursive=True ) )
        self.assertSetEqual( result, ls_result )

    def test_should_find_the_file( self ):
        result = set( find( self.root_dir, search_term='w' ) )
        result = [ a for a in result if base_name( a ) == 'w' ]
        self.assertEqual( len( result ), 1 )

    def test_should_find_files_in_folder( self ):
        result = set( find( self.root_dir, search_term='e' ) )
        result = [ a for a in result if base_name( a ) == 'e' ]
        self.assertEqual( len( result ), 1 )


class Test_find_only_files( Test_with_files ):
    defined_folders = [ 'a', 'b' ]
    defined_files = [ 'w', 'q', 'a/e', 'b/r' ]

    def test_without_search_should_be_the_same_to_ls( self ):
        result = set( find_only_files( self.root_dir ) )
        self.assertTrue( all( is_a_file( r ) for r in result ) )


class Test_find_only_folders( Test_with_files ):
    defined_folders = [ 'a', 'b' ]
    defined_files = [ 'w', 'q', 'a/e', 'b/r' ]

    def test_without_search_should_be_the_same_to_ls( self ):
        result = set( find_only_folders( self.root_dir ) )
        self.assertTrue( all( is_a_folder( r ) for r in result ) )



import tempfile
from unittest import TestCase

from faker import Factory as Faker_factory

from chibi.file import Chibi_file, Chibi_path
from chibi.file.snippets import cd
from chibi.file.temp import Chibi_temp_path


faker = Faker_factory.create()


class Test_with_files( TestCase ):
    amount_of_files = 3
    amount_of_dirs = 3
    amount_of_files_with_content = 3
    amount_of_inner_dirs = 3

    defined_folders = []
    defined_files = []

    def setUp(self):
        self.root_dir = Chibi_temp_path()
        self.empty_folder = Chibi_temp_path()
        self.folder_with_files_with_content = Chibi_temp_path()
        self.files = [  ]

        self.files = [
            self.root_dir.temp_file()
            for i in range( self.amount_of_files ) ]
        self.dirs = [
            self.root_dir.temp_dir()
            for i in range( self.amount_of_dirs ) ]

        for dir_level_1 in self.dirs:
            for i in range( self.amount_of_inner_dirs ):
                tempfile.mkdtemp( dir=str( dir_level_1 ) )

        self.files_with_content = []
        for i in range( self.amount_of_files_with_content ):
            file_path = self.folder_with_files_with_content.temp_file()
            file_path.open().write(
                faker.text(max_nb_chars=200, ext_word_list=None) )
            self.files_with_content.append( file_path )

        for f in self.defined_folders:
            ( self.root_dir + f ).mkdir()

        for f in self.defined_files:
            Chibi_file( self.root_dir + f )


class Test_moving_dir( TestCase ):
    def setUp( self ):
        self.origin_dir = Chibi_path.current_dir()

    def tearDown(self):
        cd( self.origin_dir )



from .base import Test_unit
from chibi.units.si import Metre, Gram


class Test_Metre( Test_unit ):
    def setUp( self ):
        self.unit = Metre( 10 )


class Test_Gram( Test_unit ):
    def setUp( self ):
        self.unit = Gram( 10 )



from unittest import TestCase
from chibi.object import Chibi_object
from chibi.atlas import Chibi_atlas

from chibi.module import import_, export


class Test_import( TestCase ):

    def setUp( self ):
        pass

    def test_import_chibi_object( self ):
        cls = import_( 'chibi.object.Chibi_object' )
        self.assertEqual( cls, Chibi_object )

    def test_import_chibi_atlas( self ):
        cls = import_( 'chibi.atlas.Chibi_atlas' )
        self.assertEqual( cls, Chibi_atlas )


class Test_export( TestCase ):

    def setUp( self ):
        pass

    def test_import_chibi_object( self ):
        cls = 'chibi.object.object.Chibi_object'
        self.assertEqual( cls, export( Chibi_object ) )

    def test_import_chibi_atlas( self ):
        cls = 'chibi.atlas.Chibi_atlas'
        self.assertEqual( cls, export( Chibi_atlas ) )



from chibi.file.snippets import exists
from chibi.file.temp import Chibi_temp_path
from chibi.file import Chibi_path
from tests.snippet.files import Test_with_files


class Test_path_temp( Test_with_files ):
    def test_when_is_delete_should_remove_the_folder( self ):
        path = Chibi_temp_path()
        path_str = str( path )
        self.assertTrue( exists( path ) )
        del path
        self.assertFalse( exists( path_str ) )

    def test_when_add_string_should_return_a_chibi_path( self ):
        path = Chibi_temp_path()
        result = path + "hello"
        self.assertIsInstance( result, Chibi_path )
        self.assertIn( str( path ), result )
        self.assertIn( "hello", result )

    def test_should_create_a_file( self ):
        f = self.root_dir.temp_file()
        self.assertIsInstance( f, Chibi_path )
        self.assertTrue( f.exists )

    def test_should_create_a_file_with_extencion( self ):
        f = self.root_dir.temp_file( extension='temp' )
        self.assertEqual( 'temp', f.extension )



import _string
import io
import gzip as gzip_module
import re
from collections import defaultdict

from .extra_string import codepoint, replacement
from .regex import separate_camelcase_phase_1, separate_camelcase_phase_2


__all__ = [ 'camel_to_snake' ]


def camel_to_snake( s ):
    result = separate_camelcase_phase_1.sub( r'\1_\2', s )
    result = separate_camelcase_phase_2.sub( r'\1_\2', result ).lower()
    return result


def fold_to_ascii( s ):
    if s is None:
        raise NotImplementedError
    if not isinstance( s, str ):
        raise NotImplementedError

    def none_factory():
        return None
    translate_table = codepoint + replacement
    default_translate_table = defaultdict( none_factory, translate_table )

    return s.translate( default_translate_table )


def replace_with_dict( s, d ):
    """
    replaces the substrings in the dictionary keys with their values

    Parameters
    ----------
    s: string
    d: dict

    Returns
    -------
    string

    Examples
    --------
    >>>s1 = "asdf qwer"
    >>>d1 = { 'asdf', 'foo' }
    >>>d2 = { 'qwer', 'bar' }
    >>>replace_with_dict( s1, d1)
    "foo qwer"
    >>>replace_with_dict( s1, d2)
    "asdf bar"
    """
    if s is None:
        return None
    pattern = re.compile( r'\b(' + '|'.join( d.keys() ) + r')\b' )
    return pattern.sub( lambda x: d[ x.group() ], s )


def get_the_number_of_parameters( s ):
    """
    return the number of parameters have a string

    Parameters
    ----------
    s: string

    Example
    -------
    >>>get_the_number_of_parameters( "{}" )
    1
    >>>get_the_number_of_parameters( "pi-piru {}" )
    1
    >>>get_the_number_of_parameters( "{} pi-pipu {}" )
    2
    """
    list_of_params = list( _string.formatter_parser( s ) )
    return len( list_of_params )


def decode( s, code='utf-8' ):
    '''
    decode a string
    '''
    return s.decode( code )


def gzip( s ):
    out = io.BytesIO()
    with gzip_module.GzipFile( fileobj=out, mode='w' ) as f:
        f.write( s.encode() )
    result = out.getvalue()
    return result


def remove_inner_space( s ):
    return " ".join( s.split() )


def split_in_table( s, separator_row='\n', separator=" " ):
    """
    divide una cadena con formato de tabla

    Example
    -------
    >>>table=(
        "Pipiru:     piru-piru        pipiru-pii\n"
        "Pipiru:   piru-piru pipiru-pii\n" )
    >>>split_in_table( table )
    [ 'Pipiru:', 'piru-piru', 'pipiru-pii' ],
    [ 'Pipiru:', 'piru-piru', 'pipiru-pii' ],
    [],
    """
    result = []
    rows = s.split( separator_row )
    for row in rows:
        colums = row.split( separator )
        colums = [ c for c in colums if c ]
        result.append( colums )
    return result



from unittest import TestCase

from chibi.atlas.tree import Chibi_tree, build_tree


class Value_test:
    def __init__( self, *children ):
        self.children = list( children )

    def __repr__( self ):
        return str( id( self ) )


class Test_build_tree( TestCase ):
    def setUp( self ):
        self.level_1 = [
            Value_test(), Value_test(), Value_test()
        ]

        self.level_2 = [
            Value_test( Value_test() ),
            Value_test( Value_test(), Value_test() ), Value_test()
        ]

        self.level_3 = [
            Value_test( Value_test( Value_test() ) ),
            Value_test(
                Value_test(), Value_test( *self.level_1 ) ),
            Value_test()
        ]

        a = Value_test()
        self.level_4 = [
            a, Value_test( a ), Value_test( a ),
            Value_test()
        ]

        self.child_funtion = lambda o: o.children
        self.key_funtion = lambda o: o

    def test_work_with_level_1( self ):
        result = build_tree(
            *self.level_1, key=self.key_funtion, children=self.child_funtion )
        self.assertIsInstance( result, Chibi_tree )
        self.assertEqual( len( result ), 3 )
        for r in result.values():
            self.assertEqual( len( r ), 0 )

    def test_work_with_level_2( self ):
        result = build_tree(
            *self.level_2, key=self.key_funtion, children=self.child_funtion )
        self.assertIsInstance( result, Chibi_tree )
        self.assertEqual( len( result ), 3 )
        self.assertIn(
            self.level_2[0].children[0], result[ self.level_2[0] ] )
        self.assertIn(
            self.level_2[1].children[0], result[ self.level_2[1] ] )
        self.assertIn(
            self.level_2[1].children[1], result[ self.level_2[1] ] )

        self.assertFalse( result[ self.level_2[2] ] )

    def test_work_with_level_3( self ):
        result = build_tree(
            *self.level_3, key=self.key_funtion, children=self.child_funtion )

        result_level_1 = build_tree(
            *self.level_1, key=self.key_funtion, children=self.child_funtion )
        self.assertIsInstance( result, Chibi_tree )
        self.assertEqual( len( result ), 3 )
        self.assertIn(
            self.level_3[0].children[0], result[ self.level_3[0] ] )
        self.assertIn(
            self.level_3[0].children[0].children[0],
            result[ self.level_3[0] ][ self.level_3[0].children[0] ] )

        self.assertEqual(
            result_level_1,
            result[ self.level_3[1] ][ self.level_3[1].children[1] ] )

    def test_work_with_level_3_if_be_proccess_later_level_1( self ):
        result = build_tree(
            *self.level_3, key=self.key_funtion, children=self.child_funtion )
        result_origin = build_tree(
            *self.level_3, key=self.key_funtion, children=self.child_funtion )

        result_with_level_1 = build_tree(
            *self.level_1, key=self.key_funtion, children=self.child_funtion,
            results=result )

        self.assertEqual( result_origin, result_with_level_1 )

    def test_work_with_level_4( self ):
        result = build_tree(
            *self.level_4, key=self.key_funtion, children=self.child_funtion )

        self.assertEqual(
            result[ self.level_4[1] ][ self.level_4[1].children[0] ],
            result[ self.level_4[0] ] )

        self.assertEqual(
            result[ self.level_4[1] ][ self.level_4[1].children[0] ],
            result[ self.level_4[2] ][ self.level_4[2].children[0] ] )


class Test_flip_tree( TestCase ):
    def setUp( self ):
        self.level_1 = [
            Value_test(), Value_test(), Value_test()
        ]

        self.level_3 = [
            Value_test( Value_test( Value_test() ) ),
            Value_test(
                Value_test(), Value_test( *self.level_1 ) ),
            Value_test()
        ]

        self.child_funtion = lambda o: o.children
        self.key_funtion = lambda o: o

    def test_work_with_level_4( self ):
        result = build_tree(
            *self.level_3, key=self.key_funtion, children=self.child_funtion )

        l = self.level_3
        batches = result.dependencies_batches()
        self.assertIn( l[0].children[0].children[0], batches[0] )
        self.assertIn( l[1].children[1].children[0], batches[0] )
        self.assertIn( l[1].children[1].children[1], batches[0] )
        self.assertIn( l[1].children[1].children[2], batches[0] )
        self.assertIn( l[2], batches[0] )

        self.assertIn( l[1].children[0], batches[0] )

        self.assertIn( l[0].children[0], batches[1] )
        self.assertIn( l[1].children[1], batches[1] )

        self.assertIn( l[0], batches[2] )
        self.assertIn( l[1], batches[2] )



import json
from chibi.file import Chibi_file
from chibi.atlas import _wrap


__all__ = [ 'Chibi_json' ]


class Chibi_json( Chibi_file ):
    def read( self ):
        self.reread()
        return _wrap( json.load( self.file ) )

    def write( self, data, is_safe=False ):
        super().write( json.dumps( data ) )



from . import Chibi_atlas_default
import copy


def _default_factory_tree_attr():
    return Chibi_tree()


def flat( tree, result=None ):
    if result is None:
        result = {}
    for k, v in tree.items():
        result[ k ] = v
        flat( v, result=result )
    return result


class Chibi_tree( Chibi_atlas_default ):
    def __init__( self, default_factory=None ):
        if default_factory is None:
            default_factory = _default_factory_tree_attr
        super().__init__( default_factory )

    def __repr__( self ):
        return repr( dict( self ) )

    def to_dict( self ):
        result = {}
        for k, v in self.items():
            result[k] = v.to_dict()
        return result

    def find( self, obj, recursive=True ):
        if obj in self:
            return self[ obj ]
        if recursive:
            result = None
            for v in self.values():
                result = v.find( obj, recursive=recursive )
                if result is not None:
                    return result

    def remove( self, obj, recursive=True ):
        if obj in self:
            del self[ obj ]
        if recursive:
            for v in self.values():
                v.remove( obj, recursive=recursive )

    def flat( self ):
        return flat( self )

    def dependencies_batches( self ):
        results = []
        flat_tree = flat( copy.copy( self ) )
        while flat_tree:
            current_batch = [ k for k, v in flat_tree.items() if not v ]
            if not current_batch:
                raise NotImplementedError
            for dep in current_batch:
                del flat_tree[ dep ]
                for t in flat_tree.values():
                    t.remove( dep )
            results.append( current_batch )
        return results

    def __copy__( self ):
        result = type( self )()
        for k, v in self.items():
            result[k] = copy.copy( v )
        return result


def build_tree( *objs, key=None, children=None, results=None ):
    if not key or not callable( key ):
        raise NotImplementedError
    if not children or not callable( children ):
        raise NotImplementedError

    if not results:
        results = Chibi_tree()

    for obj in objs:
        key_value = key( obj )
        children_value = children( obj )
        position_key = results.find( key_value )
        if position_key is None:
            position_key = results[ key_value ]
        for child in children_value:
            position_key[ key( child ) ]
        build_tree(
            *children_value, key=key, children=children,
            results=position_key )
    return results



from chibi.file.snippets import is_dir
from tests.snippet.files import Test_with_files


class Test_is_dir( Test_with_files ):
    amount_of_files = 3
    amount_of_dirs = 3

    def test_root_should_be_a_true( self ):
        self.assertTrue( is_dir( self.root_dir ) )

    def test_all_dirs_list_should_be_true( self ):
        for dir in self.dirs:
            self.assertTrue( is_dir( dir ) )

    def test_all_files_liist_should_be_false( self ):
        for file in self.files:
            self.assertFalse( is_dir( file ) )



from chibi.file.snippets import chown, stat
from tests.snippet.files import Test_with_files


class Test_chown_to_file( Test_with_files ):
    amount_of_files = 3
    amount_of_dirs = 3

    def test_verbose_when_no_change_the_owners( self ):
        current_stat = stat( self.files[0] )
        with self.assertLogs( level='INFO' ) as cm:
            chown( self.files[0] )
        output = cm.output[0]
        self.assertIn( 'permanece', output )
        self.assertIn(
            '{}:{}'.format(
                current_stat.user.name, current_stat.group.name ),
            output )



import tempfile

from .path import Chibi_path


class Chibi_temp_path( Chibi_path ):
    def __new__( cls, *args, **kw ):
        args_2 = []
        args_2.append( tempfile.mkdtemp() )
        return str.__new__( cls, *args_2, **kw )

    def __del__( self ):
        self.delete()

    def __add__( self, other ):
        return Chibi_path( str( self ) ) + other

    def temp_file( self, extension='' ):
        subffix = f'.{extension}' if extension else extension
        file_name = tempfile.mkstemp( suffix=subffix, dir=str( self ) )[1]
        return Chibi_path( file_name )

    def temp_dir( self ):
        return Chibi_path( tempfile.mkdtemp( dir=str( self ) ) )



import pwd
import grp
from chibi.atlas import Chibi_atlas
from chibi.file import Chibi_file
from chibi.snippet.string import split_in_table, camel_to_snake


def _parse_passwd( passwd ):
    return Chibi_atlas( dict(
        name=passwd.pw_name, passwd=passwd.pw_passwd, uid=passwd.pw_uid,
        gid=passwd.pw_gid, gecos=passwd.pw_gecos, dir=passwd.pw_dir,
        shell=passwd.pw_shell ) )


def _parse_group( group ):
    return Chibi_atlas( dict(
        name=group.gr_name, passwd=group.gr_passwd, gid=group.gr_gid,
        mem=group.gr_mem ) )


def get_passwd( uid=None, name=None ):
    if uid is not None:
        return _parse_passwd( pwd.getpwuid( uid ) )
    elif name is not None:
        return _parse_passwd( pwd.getpwnam( name ) )
    raise ValueError( "uid y name no pueden ser None" )


def get_group( gid=None, name=None ):
    if gid is not None:
        return _parse_group( grp.getgrgid( gid ) )
    elif name is not None:
        return _parse_group( grp.getgrnam( name ) )
    raise ValueError( "gid y name no pueden ser None" )


def user_exists( uid=None, name=None ):
    if uid is not None and name is None:
        return any( filter( lambda x: x.pw_uid == uid, pwd.getpwall() ) )
    elif uid is None and name is not None:
        return any( filter( lambda x: x.pw_name == name, pwd.getpwall() ) )
    else:
        return any( filter(
            lambda x: x.pw_name == name and x.pw_uid == uid, pwd.getpwall() ) )


def group_exists( gid=None, name=None ):
    if gid is not None and name is None:
        return any( filter( lambda x: x.gr_gid == gid, grp.getgrall() ) )
    elif gid is None and name is not None:
        return any( filter( lambda x: x.gr_name == name, grp.getgrall() ) )
    else:
        return any( filter(
            lambda x: x.gr_name == name and x.gr_gid == gid, grp.getgrall() ) )


def mem_info():
    f = Chibi_file( "/proc/meminfo" )
    content = f.read()
    table = split_in_table( content )
    result = Chibi_atlas()
    for row in table:
        if not row:
            continue
        key = camel_to_snake( row[0].replace( ':', '' ) )
        value = row[1]
        result[ key ] = int( value )
    return result



def guaranteed_list( d, *keys ):
    def transform( x ):
        if not x:
            result = []
        elif isinstance( x, list ):
            result = x
        else:
            result = [x]
        return result

    if isinstance( d, dict ):
        for k, v in d.items():
            if k in keys:
                d[ k ] = transform( v )
            guaranteed_list( v, *keys )
    elif isinstance( d, list ):
        for i in d:
            guaranteed_list( i, *keys )
    return d


def compress_dummy_list( d ):
    def transform( self, x ):
        if isinstance( x, list ):
            result = []
            for y in x:
                if isinstance( y, dict ) and len( y ) == 1:
                    result += list( y.values() )
                else:
                    result.append( y )
            return result
        else:
            return x

    if isinstance( d, dict ):
        for k, v in d.items():
            if isinstance( v, dict ) and len( v ) == 1:
                vk = list( v.keys() )[0]
                if vk in k:
                    d[ k ] = v[ vk ]
            compress_dummy_list( d[ k ] )
    elif isinstance( d, list ):
        for i in d:
            compress_dummy_list( i )
    return d



import os
import copy
import json
import logging
import mmap

import yaml
import magic

from chibi.atlas import _wrap
from chibi.file.path import Chibi_path
from chibi.file.snippets import (
    exists, stat, check_sum_md5, read_in_chunks, base_name, file_dir,
    copy as copy_file,
)


logger = logging.getLogger( 'chibi.file.chibi_file' )


class Chibi_file:
    def __new__( cls, path, *args, **kw ):
        from .other import find_correct_class
        cls = find_correct_class( path, cls )
        obj = super().__new__( cls )
        return obj

    def __init__( self, file_name ):
        self._file_name = file_name
        self._current_dir = file_dir( file_name )
        if not self.exists:
            self.touch()
        self.reread()

    def __str__( self ):
        return "file: '{}'".format( self.path )

    @property
    def file_name( self ):
        return base_name( self._file_name )

    @property
    def dir( self ):
        return self._current_dir

    @property
    def path( self ):
        return Chibi_path( self._file_name )

    @property
    def is_empty( self ):
        return self.properties.size == 0

    @property
    def properties( self ):
        prop = stat( self.path )
        prop.mime = magic.Magic( mime=True ).from_file( self.path )
        prop.extension = os.path.splitext( self.path )[1][1:]
        return prop

    def __del__( self ):
        try:
            self.file.close()
        except AttributeError:
            pass

    def close( self ):
        self.file.close()

    def find( self, string_to_find ):
        if isinstance( string_to_find, str ):
            string_to_find = string_to_find.encode()
        with mmap.mmap(
                self.file.fileno(), 0, prot=mmap.PROT_READ ) as f:

            return f.find( string_to_find )

    def reread( self ):
        self.file = open( self.path, 'r' )

    def __contains__( self, string ):
        return self.find( string ) >= 0

    def append( self, string ):
        if isinstance( string, ( bytes, bytearray ) ):
            with open( self.path, 'ab' ) as f:
                f.write( string )
        else:
            with open( self.path, 'a' ) as f:
                f.write( string )
        self.reread()

    def write( self, string ):
        if isinstance( string, ( bytes, bytearray ) ):
            with open( self.path, 'wb' ) as f:
                f.write( string )
        else:
            with open( self.path, 'w' ) as f:
                f.write( string )
        self.reread()

    def read( self ):
        """
        lee todo el archivo
        """
        result = self.file.read()
        self.reread()
        return result

    @property
    def file( self ):
        return self._file_content

    @file.setter
    def file( self, value ):
        try:
            old_file = self._file_content
            old_file.close()
        except AttributeError:
            pass
        self._file_content = value

    @property
    def exists( self ):
        return exists( self.path )

    def touch( self ):
        open( self.path, 'a' ).close()

    def copy( self, dest, verbose=True, **kw ):
        copy_file( self.path, dest, verbose=verbose, **kw )

    def chunk( self, chunk_size=4096 ):
        return read_in_chunks( self.path, 'r', chunk_size=chunk_size )

    def check_sum_md5( self, check_sum ):
        return check_sum_md5( self.path, check_sum )

    def read_json( self ):
        logger.warning( 'deprecated' )
        self.reread()
        return _wrap( json.load( self.file ) )

    def write_json( self, data ):
        logger.warning( 'deprecated' )
        self.write( json.dumps( data ) )

    def read_yaml( self ):
        logger.warning( 'deprecated' )
        self.reread()
        result = yaml.load( self.file, Loader=yaml.FullLoader )
        return _wrap( result )

    def write_yaml( self, data, is_safe=False ):
        logger.warning( 'deprecated' )
        if is_safe:
            self.write( yaml.safe_dump( data ) )
        else:
            self.write( yaml.dump( data ) )

    def __copy__( self ):
        return type( self )( self.path )

    def __eq__( self, other ):
        if isinstance( other, Chibi_file ):
            return self.path == other.path
        return False

    def __iadd__( self, other ):
        if ( isinstance( other, str ) ):
            self.append( other )
            return self
        if ( isinstance( other, Chibi_file ) ):
            if self == other:
                self += "".join( other.chunk() )
                return self

            other = copy.copy( other )
            for chunk in other.chunk():
                self += chunk
            return self
        self += str( other )



from tests.snippet.files import Test_with_files
from chibi.file import Chibi_file


class Test_chibi_file_append( Test_with_files ):
    def setUp( self ):
        super().setUp()
        self.chibi_file = Chibi_file( self.files_with_content[0] )
        self.chibi_file_other = Chibi_file( self.files_with_content[1] )

    def test_should_add_more_text( self ):
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertTrue( whole_file, "el archivo esta vacio" )

        self.chibi_file.append( 'explosion!!!' )
        whole_file_before = "".join( self.chibi_file.chunk() )
        self.assertNotEqual( whole_file, whole_file_before )
        self.assertGreater( whole_file_before, whole_file )

    def test_should_work_the_plus_equal__str( self ):
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertTrue( whole_file, "el archivo esta vacio" )

        self.chibi_file += 'explosion!!!'
        whole_file_before = "".join( self.chibi_file.chunk() )

        self.assertNotEqual( whole_file, whole_file_before )
        self.assertGreater( whole_file_before, whole_file )

    def test_should_work_the_plus_equal__himself( self ):
        whole_file = "".join( self.chibi_file.chunk() )
        self.assertTrue( whole_file, "el archivo esta vacio" )

        self.chibi_file += self.chibi_file
        whole_file_before = "".join( self.chibi_file.chunk() )

        self.assertNotEqual( whole_file, whole_file_before )
        self.assertEqual( whole_file * 2, whole_file_before )

    def test_should_work_the_plus_equal__another_file( self ):
        whole_file = "".join( self.chibi_file.chunk() )
        whole_file_other = "".join( self.chibi_file_other.chunk() )
        self.assertTrue( whole_file, "el archivo esta vacio" )
        self.assertTrue( whole_file_other, "el archivo esta vacio" )

        self.chibi_file += self.chibi_file_other
        whole_file_before = "".join( self.chibi_file.chunk() )

        self.assertNotEqual( whole_file, whole_file_before )
        self.assertEqual( whole_file + whole_file_other, whole_file_before )



from unittest import TestCase, skip
from chibi.file.temp import Chibi_temp_path
from chibi.file.other import Chibi_csv


class Test_chibi_csv( TestCase ):
    def setUp( self ):
        self.folder = Chibi_temp_path()
        self.file_csv = self.folder.temp_file( extension='csv' )

    def test_should_append_the_range( self ):
        csv = Chibi_csv( self.file_csv )
        csv.append( range( 10 ) )
        csv.append( range( 10 ) )
        csv.append( range( 10 ) )
        csv.append( range( 10 ) )
        result = list( csv.read_as_list() )
        self.assertTrue( result )
        self.assertEqual( len( result ), 4 )

    def test_should_append_dicts( self ):
        csv = Chibi_csv( self.file_csv )
        csv.append( { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9' } )
        csv.append(
            { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9', 'cosa4': 'a' } )
        csv.append( { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9' } )
        result = list( csv.read_as_dict() )
        self.assertEqual( len( result ), 3 )
        for r in result:
            self.assertIsInstance( r, dict  )
            self.assertTrue( r )
            self.assertEqual( { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9' }, r )

    def test_should_read_the_file_using_x_y_index( self ):
        csv = Chibi_csv( self.file_csv )
        csv.append( range( 10 ) )
        csv.append( range( 10, 20 ) )
        csv.append( range( 20, 30 ) )
        csv.append( range( 30, 40 ) )
        self.assertEqual( csv[0], list( str( i ) for i in range( 10 ) ) )
        self.assertEqual( csv[1], list( str( i ) for i in range( 10, 20  ) ) )
        self.assertEqual( csv[2], list( str( i ) for i in range( 20, 30  ) ) )
        self.assertEqual( csv[3], list( str( i ) for i in range( 30, 40  ) ) )

    def test_should_raise_a_exception_when_out_of_range( self ):
        csv = Chibi_csv( self.file_csv )
        csv.append( range( 10 ) )
        with self.assertRaises( IndexError ):
            csv[1]

    @skip( (
        'esta prueba esta incompleta porque cambiara el '
        'funcionamiento de los headers' ) )
    def test_when_append_dict_should_be_marked_like_has_headers( self ):
        self.fail( 'incomplreto' )
        csv = Chibi_csv( self.file_csv )
        self.assertFalse( csv.has_headers )
        csv.append( { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9' } )
        csv.append(
            { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9', 'cosa4': 'a' } )
        csv.append( { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9' } )
        result = list( csv.read_as_dict() )
        self.assertEqual( len( result ), 3 )
        for r in result:
            self.assertIsInstance( r, dict  )
            self.assertTrue( r )
            self.assertEqual( { 'cosa1': '1', 'cosa2': '3', 'cosa3': '9' }, r )



import os
from .chibi_csv import Chibi_csv
from .chibi_yaml import Chibi_yaml
from .chibi_json import Chibi_json
from .chibi_python import Chibi_python
from .chibi_systemd import Chibi_systemd  # noqa: F401
from .conf import Chibi_conf  # noqa: F401


__all__ = [ 'Chibi_csv', 'Chibi_yaml', 'Chibi_json', 'Chibi_python' ]


def find_correct_class( path, cls ):
    file_name, ext = os.path.splitext( path )
    if ext == '.csv':
        return Chibi_csv
    elif ext in ( '.yml', '.yaml' ):
        return Chibi_yaml
    elif ext == '.json':
        return Chibi_json
    elif ext == '.py':
        return Chibi_python
    return cls
